from __future__ import annotations

from typing import TYPE_CHECKING, Any

import polars as pl

import xorq.vendor.ibis.expr.datatypes as dt
from xorq.vendor.ibis.expr.schema import Schema
from xorq.vendor.ibis.formats import DataMapper, SchemaMapper, TableProxy, TypeMapper


if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd
    import pyarrow as pa


_to_polars_types = {
    dt.Boolean: pl.Boolean,
    dt.Null: pl.Null,
    dt.String: pl.Utf8,
    dt.Binary: pl.Binary,
    dt.Date: pl.Date,
    dt.Time: pl.Time,
    dt.Int8: pl.Int8,
    dt.Int16: pl.Int16,
    dt.Int32: pl.Int32,
    dt.Int64: pl.Int64,
    dt.UInt8: pl.UInt8,
    dt.UInt16: pl.UInt16,
    dt.UInt32: pl.UInt32,
    dt.UInt64: pl.UInt64,
    dt.Float32: pl.Float32,
    dt.Float64: pl.Float64,
}

_from_polars_types = {v: k for k, v in _to_polars_types.items()}


class PolarsType(TypeMapper):
    @classmethod
    def to_ibis(cls, typ: pl.DataType, nullable=True) -> dt.DataType:
        """Convert a polars type to an ibis type."""

        base_type = typ.base_type()
        if base_type in (pl.Categorical, pl.Enum):
            return dt.String(nullable=nullable)
        elif base_type is pl.Decimal:
            return dt.Decimal(
                precision=typ.precision, scale=typ.scale, nullable=nullable
            )
        elif base_type is pl.Datetime:
            try:
                timezone = typ.time_zone
            except AttributeError:  # pragma: no cover
                timezone = typ.tz  # pragma: no cover

            # this will raise on polars for seconds "s" unit as it's not supported
            return dt.Timestamp.from_unit(
                unit=typ.time_unit, timezone=timezone, nullable=nullable
            )

        elif base_type is pl.Duration:
            try:
                time_unit = typ.time_unit
            except AttributeError:  # pragma: no cover
                time_unit = typ.tu  # pragma: no cover
            return dt.Interval(unit=time_unit, nullable=nullable)
        elif base_type is pl.List:
            return dt.Array(cls.to_ibis(typ.inner), nullable=nullable)
        elif base_type is pl.Struct:
            return dt.Struct.from_tuples(
                [(field.name, cls.to_ibis(field.dtype)) for field in typ.fields],
                nullable=nullable,
            )
        else:
            return _from_polars_types[base_type](nullable=nullable)

    @classmethod
    def from_ibis(cls, dtype: dt.DataType) -> pl.DataType:
        """Convert an ibis type to a polars type."""
        if dtype.is_decimal():
            return pl.Decimal(
                precision=dtype.precision,
                scale=9 if dtype.scale is None else dtype.scale,
            )
        elif dtype.is_timestamp():
            unit = dtype.unit.short
            if unit in {"us", "ns", "ms"}:
                return pl.Datetime(unit, dtype.timezone)
            else:
                # this for "s", if something else is passed, it'll raise at
                # the from_unit level.
                return pl.Datetime("ns", dtype.timezone)  # this was the default before

        elif dtype.is_interval():
            if dtype.unit.short in {"us", "ns", "ms"}:
                return pl.Duration(dtype.unit.short)
            else:
                raise ValueError(f"Unsupported polars duration unit: {dtype.unit}")
        elif dtype.is_struct():
            fields = [
                pl.Field(name=name, dtype=cls.from_ibis(dtype))
                for name, dtype in dtype.fields.items()
            ]
            return pl.Struct(fields)
        elif dtype.is_array():
            return pl.List(cls.from_ibis(dtype.value_type))
        else:
            try:
                return _to_polars_types[type(dtype)]
            except KeyError:
                raise NotImplementedError(
                    f"Converting {dtype} to polars is not supported yet"
                )


class PolarsSchema(SchemaMapper):
    @classmethod
    def from_ibis(cls, schema: Schema) -> dict[str, pl.DataType]:
        """Convert a schema to a polars schema."""
        return {name: PolarsType.from_ibis(typ) for name, typ in schema.items()}

    @classmethod
    def to_ibis(cls, schema: dict[str, pl.DataType]) -> Schema:
        """Convert a polars schema to a schema."""
        return Schema.from_tuples(
            [(name, PolarsType.to_ibis(typ)) for name, typ in schema.items()]
        )


class PolarsData(DataMapper):
    @classmethod
    def infer_scalar(cls, scalar: Any) -> dt.DataType:
        """Infer the ibis type of a scalar."""
        return PolarsType.to_ibis(pl.Series(values=[scalar]).dtype)

    @classmethod
    def infer_column(cls, column: Sequence) -> dt.DataType:
        """Infer the ibis type of a sequence."""
        if not isinstance(column, pl.Series):
            column = pl.Series(values=column)
        return PolarsType.to_ibis(column.dtype)

    @classmethod
    def infer_table(cls, table) -> Schema:
        """Infer the schema of a table."""
        if not isinstance(table, pl.DataFrame):
            table = pl.DataFrame(table)

        return PolarsSchema.to_ibis(table.schema)

    @classmethod
    def convert_scalar(cls, df: pl.DataFrame, dtype: dt.DataType) -> Any:
        assert df.shape == (1, 1)
        df = cls.convert_table(df, Schema({df.columns[0]: dtype}))
        return df[0, 0]

    @classmethod
    def convert_column(cls, df: pl.DataFrame, dtype: dt.DataType) -> pl.Series:
        assert df.shape[1] == 1
        df = cls.convert_table(df, Schema({df.columns[0]: dtype}))
        return df[:, 0]

    @classmethod
    def convert_table(cls, df: pl.DataFrame, schema: Schema) -> pl.DataFrame:
        pl_schema = PolarsSchema.from_ibis(schema)

        if df.schema == pl_schema:
            return df
        return df.cast(pl_schema)


class PolarsDataFrameProxy(TableProxy[pl.DataFrame]):
    def proxy_equals(self, other: "PolarsDataFrameProxy") -> bool:
        return self.obj.frame_equal(other.obj)

    def to_frame(self) -> pd.DataFrame:
        return self.obj.to_pandas()

    def to_pyarrow(self, schema: Schema) -> pa.Table:
        from xorq.vendor.ibis.formats.pyarrow import PyArrowData

        table = self.obj.to_arrow()
        return PyArrowData.convert_table(table, schema)

    def to_polars(self, schema: Schema) -> pl.DataFrame:
        return self.obj



import pyarrow as pa
import pytest
from pandas.testing import assert_frame_equal

from xorq.expr.translate import plan_to_ibis
from xorq.internal import ContextProvider
from xorq.sql import parser
from xorq.vendor import ibis


@pytest.fixture(scope="session")
def con():
    return ibis.connect("duckdb://")


@pytest.fixture(scope="session")
def t(con):
    con.create_table(
        "t",
        pa.Table.from_pydict(
            {
                "a": ["a1", "a2", "a3", "a4", "a5", "a6", "a7"],
                "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
                "c": [1, 2, 3, 4, 5, 6, 7],
                "d": [5, 6, 1, 7, 2, 4, 3],
                "e": [1, 2, 3, 4, 5, 6, 7],
            }
        ),
    )

    return con.table("t")


@pytest.fixture(scope="session")
def s(con):
    con.create_table(
        "s",
        pa.Table.from_pydict(
            {
                "a": ["a1", "a2", "a3", "a4", "a5", "a9", "a8"],
                "f": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
                "g": [True, False, False, False, False, True, True],
            }
        ),
    )

    return con.table("s")


@pytest.mark.parametrize(
    "sql",
    [
        "select t.a, t.b from t where t.c > 2",
        "select t.a, t.b from t where t.c > 2 limit 3",
        "select t.a, t.b * 2.5 from t",
        "select t.a, t.b / 2.5 from t",
    ],
)
def test_table_scan(con, sql, t):
    # from sql to LogicalPlan
    plan = parser.parse_sql(sql, ContextProvider({"t": t.schema().to_pyarrow()}))

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"t": t.schema()})

    expected = con.table("t").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)


@pytest.mark.parametrize(
    "sql",
    [
        "select s.a from s where s.g is true",
        "select s.f from s where s.g is false",
        "select s.f from s where s.g is null",
        "select s.f from s where s.g is not null",
    ],
)
def test_table_scan_bool(con, sql, s):
    # from sql to LogicalPlan
    plan = parser.parse_sql(sql, ContextProvider({"s": s.schema().to_pyarrow()}))

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"s": s.schema()})

    expected = con.table("s").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)


@pytest.mark.xfail(reason="datafusion 42.0.0 update introduced a bug")
@pytest.mark.parametrize(
    "sql",
    [
        "select t.a, t.b from t order by t.d",
        "select t.a, t.b from t order by t.d limit 5",
    ],
)
def test_sort(con, sql, t):
    # from sql to LogicalPlan
    plan = parser.parse_sql(sql, ContextProvider({"t": t.schema().to_pyarrow()}))

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"t": t.schema()})

    expected = con.table("t").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)


@pytest.mark.xfail(reason="datafusion 51.0.0 update introduced a bug")
def test_agg(con, t):
    sql = "select sum(t.b) from t"

    # from sql to LogicalPlan
    plan = parser.parse_sql(sql, ContextProvider({"t": t.schema().to_pyarrow()}))

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"t": t.schema()})

    expected = con.table("t").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)


@pytest.mark.parametrize(
    "sql",
    [
        "select s.a from (select t.a from t) as s",
    ],
)
def test_subquery_alias(con, sql, t):
    # from sql to LogicalPlan
    plan = parser.parse_sql(sql, ContextProvider({"t": t.schema().to_pyarrow()}))

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"t": t.schema()})

    expected = con.table("t").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)


@pytest.mark.parametrize(
    "sql",
    [
        "select t.a, s.f + t.b from t join s on t.a = s.a",
        "select t.a, s.f - t.b from t left join s on t.a = s.a",
        "select t.a, s.f - t.b from t right join s on t.a = s.a",
        "select t.a, t.b, s.f from t join s on t.a = s.a where t.b > 1.5 limit 2",
    ],
)
def test_join(con, sql, t, s):
    # from sql to LogicalPlan
    plan = parser.parse_sql(
        sql,
        ContextProvider({"t": t.schema().to_pyarrow(), "s": s.schema().to_pyarrow()}),
    )

    # from LogicalPlan back to ibis
    expr = plan_to_ibis(plan, {"t": t.schema(), "s": s.schema()})

    expected = con.table("t").sql(sql).execute()
    actual = con.execute(expr)
    actual.columns = expected.columns

    assert expr is not None
    assert_frame_equal(expected, actual)



import datetime
import decimal

import pytest

import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.common import (
    RefEnum,
    RegistryEnum,
)
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_unbound_table(t, compiler):
    yaml_dict = compiler.to_yaml(t)
    node_ref = yaml_dict["expression"][RefEnum.node_ref]
    expression = yaml_dict["definitions"][RegistryEnum.nodes][node_ref]
    assert expression["op"] == "UnboundTable"
    assert expression["name"] == "test_table"
    assert expression[RefEnum.schema_ref]

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.schema() == t.schema()
    assert roundtrip_expr.op().name == t.op().name


def test_field(t, compiler):
    expr = t.a
    yaml_dict = compiler.to_yaml(expr)

    expression = yaml_dict["expression"]
    assert expression["op"] == "Field"
    assert expression["name"] == "a"
    assert get_dtype_yaml(yaml_dict, expression) == {
        "op": "DataType",
        "type": "Int64",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)
    assert roundtrip_expr.get_name() == expr.get_name()


def test_literal(compiler):
    lit = ibis.literal(42)
    yaml_dict = compiler.to_yaml(lit)

    expression = yaml_dict["expression"]
    assert expression["op"] == "Literal"
    assert expression["value"] == 42
    dtype_yaml = yaml_dict["definitions"][RegistryEnum.dtypes][
        expression["type"][RefEnum.dtype_ref]
    ]
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Int8",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(lit)


def test_binary_op(t, compiler):
    expr = t.a + 1
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Add"
    assert expression["right"]["op"] == "Literal"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


@pytest.mark.parametrize(
    "primitive",
    [
        (ibis.literal(True), "Boolean"),
        (ibis.literal(1), "Int8"),
        (ibis.literal(1000), "Int16"),
        (ibis.literal(1.0), "Float64"),
        (ibis.literal("hello"), "String"),
        (ibis.literal(None), "Null"),
    ],
)
def test_primitive_types(compiler, primitive):
    lit, expected_type = primitive
    yaml_dict = compiler.to_yaml(lit)

    expression = yaml_dict["expression"]
    assert expression["op"] == "Literal"
    assert get_dtype_yaml(yaml_dict, expression)["type"] == expected_type

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(lit)
    assert roundtrip_expr.type().name == lit.type().name


def test_temporal_types(compiler):
    now = datetime.datetime.now()
    today = datetime.date.today()
    time = datetime.time(12, 0)
    temporals = [
        (ibis.literal(now), "Timestamp"),
        (ibis.literal(today), "Date"),
        (ibis.literal(time), "Time"),
    ]
    for lit, expected_type in temporals:
        yaml_dict = compiler.to_yaml(lit)
        expression = yaml_dict["expression"]
        assert expression["op"] == "Literal"
        assert get_dtype_yaml(yaml_dict, expression)["type"] == expected_type

        roundtrip_expr = compiler.from_yaml(yaml_dict)
        assert roundtrip_expr.equals(lit)
        assert roundtrip_expr.type().name == lit.type().name


def test_decimal_type(compiler):
    dec = decimal.Decimal("123.45")
    lit = ibis.literal(dec)
    yaml_dict = compiler.to_yaml(lit)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Literal"
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml["type"] == "Decimal"
    assert dtype_yaml["nullable"]

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(lit)
    assert roundtrip_expr.type().name == lit.type().name


def test_array_type(compiler):
    lit = ibis.literal([1, 2, 3])
    yaml_dict = compiler.to_yaml(lit)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    inner = get_dtype_yaml(yaml_dict, dtype_yaml, key="value_type")
    assert expression["op"] == "Literal"
    assert expression["value"] == (1, 2, 3)
    assert dtype_yaml["type"] == "Array"
    assert inner["type"] == "Int8"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(lit)
    assert roundtrip_expr.type().value_type == lit.type().value_type


def test_map_type(compiler):
    lit = ibis.literal({"a": 1, "b": 2})
    yaml_dict = compiler.to_yaml(lit)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    key_dtype_yaml = get_dtype_yaml(yaml_dict, dtype_yaml, key="key_type")
    value_dtype_yaml = get_dtype_yaml(yaml_dict, dtype_yaml, key="value_type")
    assert expression["op"] == "Literal"
    assert dtype_yaml["type"] == "Map"
    assert key_dtype_yaml["type"] == "String"
    assert value_dtype_yaml["type"] == "Int8"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(lit)
    assert roundtrip_expr.type().key_type == lit.type().key_type
    assert roundtrip_expr.type().value_type == lit.type().value_type


def test_complex_expression_roundtrip(t, compiler):
    expr = (t.a + 1).abs() * 2
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_window_function_roundtrip(t, compiler):
    expr = t.a.sum().over(ibis.window(group_by=t.a))
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_join_roundtrip(t, compiler):
    t2 = ibis.table({"b": "int64"}, name="test_table_2")
    expr = t.join(t2, t.a == t2.b)
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.schema() == expr.schema()


def test_aggregation_roundtrip(t, compiler):
    expr = t.group_by(t.a).aggregate(count=t.a.count())
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.schema() == expr.schema()


def test_aggregation_with_alias_roundtrip(t, compiler):
    expr = t.group_by(t.a.name("cust_id")).aggregate(count=t.a.count())
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.schema() == expr.schema()


def test_table_perserves_namespace(compiler):
    expr = ibis.table({"b": "int64"}, name="t2", database="db", catalog="catalog")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.op().namespace == expr.op().namespace



"""Tests for deferred_cross_val_score."""

import numpy as np
import pytest
from pytest import param

import xorq.api as xo


sklearn = pytest.importorskip("sklearn")

from sklearn.model_selection import (  # noqa: E402
    KFold,
    RepeatedKFold,
    RepeatedStratifiedKFold,
    ShuffleSplit,
    StratifiedKFold,
    StratifiedShuffleSplit,
    TimeSeriesSplit,
    cross_val_score,
)

from xorq.expr.ml.cross_validation import (  # noqa: E402
    CrossValScore,
    apply_deterministic_sort,
    deferred_cross_val_score,
)


@pytest.fixture(scope="module")
def classification_data():
    """Generate a classification dataset as an ibis table."""
    import pandas as pd
    from sklearn.datasets import make_classification

    X, y = make_classification(
        n_samples=200,
        n_features=4,
        n_informative=3,
        n_redundant=0,
        random_state=42,
    )
    df = pd.DataFrame(X, columns=[f"f{i}" for i in range(4)]).assign(
        target=y, t=range(200)
    )
    return xo.memtable(df)


@pytest.fixture(scope="module")
def regression_data():
    """Generate a regression dataset as an ibis table."""
    import pandas as pd
    from sklearn.datasets import make_regression

    X, y = make_regression(
        n_samples=200,
        n_features=4,
        n_informative=3,
        random_state=42,
    )
    df = pd.DataFrame(X, columns=[f"f{i}" for i in range(4)]).assign(
        target=y, t=range(200)
    )
    return xo.memtable(df)


@pytest.fixture(scope="module")
def classifier_pipeline():
    """An unfitted xorq Pipeline for classification."""
    from sklearn.linear_model import LogisticRegression
    from sklearn.pipeline import Pipeline as SklearnPipeline
    from sklearn.preprocessing import StandardScaler

    from xorq.expr.ml.pipeline_lib import Pipeline

    return Pipeline.from_instance(
        SklearnPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42, max_iter=1000)),
            ]
        )
    )


@pytest.fixture(scope="module")
def regressor_pipeline():
    """An unfitted xorq Pipeline for regression."""
    from sklearn.linear_model import LinearRegression
    from sklearn.pipeline import Pipeline as SklearnPipeline
    from sklearn.preprocessing import StandardScaler

    from xorq.expr.ml.pipeline_lib import Pipeline

    return Pipeline.from_instance(
        SklearnPipeline(
            [
                ("scaler", StandardScaler()),
                ("reg", LinearRegression()),
            ]
        )
    )


FEATURES = tuple(f"f{i}" for i in range(4))
TARGET = "target"
RANDOM_SEED = 42


# --- splitter instances, each with its expected number of folds ---
CLASSIFIER_SPLITTERS = (
    param(
        KFold(n_splits=3, shuffle=True, random_state=42),
        3,
        id="KFold",
    ),
    param(
        StratifiedKFold(n_splits=5, shuffle=True, random_state=42),
        5,
        id="StratifiedKFold",
    ),
    param(
        RepeatedKFold(n_splits=3, n_repeats=2, random_state=42),
        6,
        id="RepeatedKFold",
    ),
    param(
        RepeatedStratifiedKFold(n_splits=3, n_repeats=2, random_state=42),
        6,
        id="RepeatedStratifiedKFold",
    ),
    param(
        ShuffleSplit(n_splits=4, test_size=0.25, random_state=42),
        4,
        id="ShuffleSplit",
    ),
    param(
        StratifiedShuffleSplit(n_splits=4, test_size=0.25, random_state=42),
        4,
        id="StratifiedShuffleSplit",
    ),
)


# --- Tests with integer cv (uses train_test_splits internally) ---


def test_int_cv_returns_cross_val_score(classification_data, classifier_pipeline):
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )
    assert isinstance(result, CrossValScore)
    scores = result.execute()
    assert isinstance(scores, np.ndarray)


def test_int_cv_correct_number_of_folds(classification_data, classifier_pipeline):
    for k in (3, 5):
        result = deferred_cross_val_score(
            classifier_pipeline,
            classification_data,
            features=FEATURES,
            target=TARGET,
            cv=k,
            random_seed=42,
        )
        assert len(result) == k
        assert len(result.execute()) == k


def test_int_cv_scores_are_finite(classification_data, classifier_pipeline):
    scores = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    ).execute()
    assert np.all(np.isfinite(scores))


def test_int_cv_classifier_scores_bounded(classification_data, classifier_pipeline):
    """Default classifier scorer is accuracy, which is in [0, 1]."""
    scores = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    ).execute()
    assert np.all(scores >= 0.0)
    assert np.all(scores <= 1.0)


def test_int_cv_regressor_default_scorer(regression_data, regressor_pipeline):
    scores = deferred_cross_val_score(
        regressor_pipeline,
        regression_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    ).execute()
    assert isinstance(scores, np.ndarray)
    assert len(scores) == 3
    assert np.all(np.isfinite(scores))


def test_int_cv_explicit_scorer(classification_data, classifier_pipeline):
    scores = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        scoring="f1",
        random_seed=42,
    ).execute()
    assert isinstance(scores, np.ndarray)
    assert len(scores) == 3
    assert np.all(np.isfinite(scores))


def test_int_cv_mean_close_to_sklearn(classification_data, classifier_pipeline):
    """int cv uses hash-based folds (different from sklearn's KFold), so
    per-fold scores differ.  But the mean over k folds should converge to
    the same value.  We allow 5% tolerance on the mean."""
    df = classification_data.execute()
    X = df[list(FEATURES)].values
    y = df[TARGET].values

    for k in (3, 5):
        xorq_scores = deferred_cross_val_score(
            classifier_pipeline,
            classification_data,
            features=FEATURES,
            target=TARGET,
            cv=k,
            random_seed=42,
        ).execute()
        sklearn_scores = cross_val_score(
            classifier_pipeline.instance,
            X,
            y,
            cv=KFold(n_splits=k, shuffle=True, random_state=42),
            scoring="accuracy",
        )
        assert abs(xorq_scores.mean() - sklearn_scores.mean()) < 0.05


def test_int_cv_reproducible_with_seed(classification_data, classifier_pipeline):
    scores1 = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    ).execute()
    scores2 = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    ).execute()
    np.testing.assert_array_equal(scores1, scores2)


# --- Tests that each sklearn splitter produces the right shape and finite scores ---


@pytest.mark.parametrize("cv, expected_folds", CLASSIFIER_SPLITTERS)
def test_splitter_produces_correct_folds(
    classification_data,
    classifier_pipeline,
    cv,
    expected_folds,
):
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=cv,
    )
    assert len(result) == expected_folds
    scores = result.execute()
    assert isinstance(scores, np.ndarray)
    assert len(scores) == expected_folds
    assert np.all(np.isfinite(scores))


# --- Per-fold parity: deferred_cross_val_score must match sklearn exactly ---
#
# When both sides use the same splitter, they see identical train/test
# indices, so scores must be numerically equal.


@pytest.mark.parametrize(
    "make_cv, scoring",
    (
        param(
            lambda: KFold(n_splits=3, shuffle=True, random_state=42),
            "accuracy",
            id="KFold-accuracy",
        ),
        param(
            lambda: StratifiedKFold(n_splits=5, shuffle=True, random_state=42),
            "f1",
            id="StratifiedKFold-f1",
        ),
        param(
            lambda: RepeatedKFold(n_splits=3, n_repeats=2, random_state=42),
            "accuracy",
            id="RepeatedKFold-accuracy",
        ),
        param(
            lambda: RepeatedStratifiedKFold(n_splits=3, n_repeats=2, random_state=42),
            "f1",
            id="RepeatedStratifiedKFold-f1",
        ),
        param(
            lambda: ShuffleSplit(n_splits=4, test_size=0.25, random_state=42),
            "accuracy",
            id="ShuffleSplit-accuracy",
        ),
        param(
            lambda: StratifiedShuffleSplit(n_splits=4, test_size=0.25, random_state=42),
            "accuracy",
            id="StratifiedShuffleSplit-accuracy",
        ),
    ),
)
def test_classifier_matches_sklearn(
    classification_data,
    classifier_pipeline,
    make_cv,
    scoring,
):
    xorq_scores = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=make_cv(),
        scoring=scoring,
        random_seed=RANDOM_SEED,
    ).execute()

    # Sort by the same deterministic key so sklearn sees identical row order.
    df = apply_deterministic_sort(
        classification_data, random_seed=RANDOM_SEED
    ).execute()
    sklearn_scores = cross_val_score(
        classifier_pipeline.instance,
        df[list(FEATURES)].values,
        df[TARGET].values,
        cv=make_cv(),
        scoring=scoring,
    )

    np.testing.assert_allclose(xorq_scores, sklearn_scores, rtol=1e-9, atol=1e-12)


def test_timeseries_split_matches_sklearn(classification_data, classifier_pipeline):
    """TimeSeriesSplit with order_by must match sklearn on sorted data."""
    make_cv = lambda: TimeSeriesSplit(n_splits=4)  # noqa: E731

    xorq_scores = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=make_cv(),
        scoring="accuracy",
        order_by="t",
    ).execute()

    # sklearn sees the same order: sorted by "t".
    df = classification_data.execute().sort_values("t")
    sklearn_scores = cross_val_score(
        classifier_pipeline.instance,
        df[list(FEATURES)].values,
        df[TARGET].values,
        cv=make_cv(),
        scoring="accuracy",
    )

    np.testing.assert_allclose(xorq_scores, sklearn_scores, rtol=1e-9, atol=1e-12)


def test_regressor_kfold_matches_sklearn(regression_data, regressor_pipeline):
    """Per-fold r2 with KFold must match sklearn exactly."""
    make_cv = lambda: KFold(n_splits=3, shuffle=True, random_state=42)  # noqa: E731

    xorq_scores = deferred_cross_val_score(
        regressor_pipeline,
        regression_data,
        features=FEATURES,
        target=TARGET,
        cv=make_cv(),
        random_seed=RANDOM_SEED,
    ).execute()

    df = apply_deterministic_sort(regression_data, random_seed=RANDOM_SEED).execute()
    sklearn_scores = cross_val_score(
        regressor_pipeline.instance,
        df[list(FEATURES)].values,
        df[TARGET].values,
        cv=make_cv(),
        scoring="r2",
    )

    np.testing.assert_allclose(xorq_scores, sklearn_scores, rtol=1e-9, atol=1e-12)


# --- Tests that expression construction is fully deferred (no eager execution) ---
#
# We patch ``Expr.execute`` at the base class so that *any* ``.execute()``
# call on *any* expression (not just the top-level input) is detected.
# The previous tests only patched the input table, which missed internal
# executions such as ``FittedStep.deferred_model.execute()``.


def test_sklearn_splitter_no_execute_during_construction(
    classification_data,
    classifier_pipeline,
    monkeypatch,
):
    """Building a CrossValScore with an sklearn splitter must NOT call
    Expr.execute() anywhere in the expression graph."""
    from xorq.vendor.ibis.expr.types import Expr

    def raise_error(*args, **kwargs):
        raise AssertionError("Expr.execute() was called during expression construction")

    monkeypatch.setattr(Expr, "execute", raise_error)

    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=KFold(n_splits=3, shuffle=True, random_state=42),
    )

    assert isinstance(result, CrossValScore)
    assert len(result) == 3


def test_int_cv_no_execute_during_construction(
    classification_data,
    classifier_pipeline,
    monkeypatch,
):
    """Building a CrossValScore with int cv must NOT call
    Expr.execute() anywhere in the expression graph."""
    from xorq.vendor.ibis.expr.types import Expr

    def raise_error(*args, **kwargs):
        raise AssertionError("Expr.execute() was called during expression construction")

    monkeypatch.setattr(Expr, "execute", raise_error)

    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )

    assert isinstance(result, CrossValScore)
    assert len(result) == 3


# --- Tests for the fold_expr attribute on CrossValScore ---


def test_fold_expr_has_fold_columns_int_cv(classification_data, classifier_pipeline):
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )
    fold_df = result.fold_expr.execute()
    for i in range(3):
        assert f"fold_{i}" in fold_df.columns
    # Original columns should still be present
    for col in list(FEATURES) + [TARGET]:
        assert col in fold_df.columns


def test_fold_expr_has_fold_columns_sklearn_cv(
    classification_data, classifier_pipeline
):
    cv = KFold(n_splits=5, shuffle=True, random_state=42)
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=cv,
        random_seed=42,
    )
    fold_df = result.fold_expr.execute()
    for i in range(5):
        assert f"fold_{i}" in fold_df.columns


def test_fold_expr_values_are_0_1_or_2(classification_data, classifier_pipeline):
    """Fold columns use ternary encoding: 0=unused, 1=train, 2=test."""
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )
    fold_df = result.fold_expr.execute()
    for i in range(3):
        values = set(fold_df[f"fold_{i}"].unique())
        assert values <= {0, 1, 2}


def test_fold_expr_each_row_is_test_in_exactly_one_fold_int_cv(
    classification_data, classifier_pipeline
):
    """With int cv (equal-sized folds), each row should be test in exactly one fold."""
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )
    fold_df = result.fold_expr.execute()
    # Each row is test (2) in exactly one fold
    test_counts = sum((fold_df[f"fold_{i}"] == 2).astype(int) for i in range(3))
    assert (test_counts == 1).all()


def test_fold_expr_row_count_matches_original(classification_data, classifier_pipeline):
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=3,
        random_seed=42,
    )
    assert result.fold_expr.count().execute() == classification_data.count().execute()


@pytest.mark.parametrize("cv, expected_folds", CLASSIFIER_SPLITTERS)
def test_fold_expr_sklearn_splitter_columns(
    classification_data, classifier_pipeline, cv, expected_folds
):
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=cv,
    )
    fold_df = result.fold_expr.execute()
    for i in range(expected_folds):
        assert f"fold_{i}" in fold_df.columns
        values = set(fold_df[f"fold_{i}"].unique())
        assert values <= {0, 1, 2}


def test_fold_expr_timeseries_split(classification_data, classifier_pipeline):
    """TimeSeriesSplit fold_expr uses ternary encoding with unused rows."""
    result = deferred_cross_val_score(
        classifier_pipeline,
        classification_data,
        features=FEATURES,
        target=TARGET,
        cv=TimeSeriesSplit(n_splits=4),
        order_by="t",
    )
    fold_df = result.fold_expr.execute()
    for i in range(4):
        assert f"fold_{i}" in fold_df.columns
        values = set(fold_df[f"fold_{i}"].unique())
        assert values <= {0, 1, 2}
        # TimeSeriesSplit has unused rows in early folds
        if i < 3:
            assert 0 in values, f"fold_{i} should have unused rows"


# --- Tests for input validation ---


def test_rejects_non_pipeline(classification_data):
    with pytest.raises(TypeError, match="pipeline must be a Pipeline instance"):
        deferred_cross_val_score(
            "not_a_pipeline",
            classification_data,
            features=FEATURES,
            target=TARGET,
        )


@pytest.mark.parametrize(
    "cv",
    (
        param(
            pytest.importorskip("sklearn.model_selection").GroupKFold(n_splits=3),
            id="GroupKFold",
        ),
        param(
            pytest.importorskip("sklearn.model_selection").GroupShuffleSplit(
                n_splits=3, random_state=42
            ),
            id="GroupShuffleSplit",
        ),
        param(
            pytest.importorskip("sklearn.model_selection").LeaveOneGroupOut(),
            id="LeaveOneGroupOut",
        ),
        param(
            pytest.importorskip("sklearn.model_selection").StratifiedGroupKFold(
                n_splits=3
            ),
            id="StratifiedGroupKFold",
        ),
    ),
)
def test_rejects_group_based_splitters(classification_data, classifier_pipeline, cv):
    with pytest.raises(TypeError, match="Group-based splitters are not supported"):
        deferred_cross_val_score(
            classifier_pipeline,
            classification_data,
            features=FEATURES,
            target=TARGET,
            cv=cv,
        )


def test_rejects_timeseries_split_without_order_by(
    classification_data, classifier_pipeline
):
    with pytest.raises(TypeError, match="TimeSeriesSplit requires order_by"):
        deferred_cross_val_score(
            classifier_pipeline,
            classification_data,
            features=FEATURES,
            target=TARGET,
            cv=TimeSeriesSplit(n_splits=4),
        )



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from xorq.vendor.ibis.common.bases import Abstract
from xorq.vendor.ibis.common.grounds import Concrete
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001


if TYPE_CHECKING:
    from collections.abc import Sequence

    import xorq.vendor.ibis.expr.types as ir


class Expandable(Abstract):
    __slots__ = ()

    @abc.abstractmethod
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        """Expand `table` into value expressions that match the selector.

        Parameters
        ----------
        table
            An ibis table expression

        Returns
        -------
        Sequence[Value]
            A sequence of value expressions that match the selector

        """


class Selector(Concrete, Expandable):
    """A column selector."""

    @abc.abstractmethod
    def expand_names(self, table: ir.Table) -> frozenset[str]:
        """Compute the set of column names that match the selector."""

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        names = self.expand_names(table)
        return list(map(table.__getitem__, filter(names.__contains__, table.columns)))

    def __and__(self, other: Selector) -> Selector:
        """Compute the logical conjunction of two `Selector`s.

        Parameters
        ----------
        other
            Another selector
        """
        if not isinstance(other, Selector):
            return NotImplemented
        return And(self, other)

    def __or__(self, other: Selector) -> Selector:
        """Compute the logical disjunction of two `Selector`s.

        Parameters
        ----------
        other
            Another selector
        """
        if not isinstance(other, Selector):
            return NotImplemented
        return Or(self, other)

    def __invert__(self) -> Selector:
        """Compute the logical negation of a `Selector`."""
        return Not(self)


class Or(Selector):
    left: Selector
    right: Selector

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return self.left.expand_names(table) | self.right.expand_names(table)


class And(Selector):
    left: Selector
    right: Selector

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return self.left.expand_names(table) & self.right.expand_names(table)


class Any(Selector):
    selectors: VarTuple[Selector]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        names = (selector.expand_names(table) for selector in self.selectors)
        return frozenset.union(*names)


class All(Selector):
    selectors: VarTuple[Selector]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        names = (selector.expand_names(table) for selector in self.selectors)
        return frozenset.intersection(*names)


class Not(Selector):
    selector: Selector

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        names = self.selector.expand_names(table)
        return frozenset(col for col in table.columns if col not in names)



import functools
import itertools
import pathlib

import dask
import pandas as pd
import pytest
from attr import (
    field,
    frozen,
)
from attr.validators import (
    in_,
    optional,
)

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
)
from xorq.common.utils.defer_utils import (
    deferred_read_csv,
    deferred_read_parquet,
)
from xorq.common.utils.inspect_utils import (
    get_partial_arguments,
)
from xorq.tests.util import assert_frame_equal


duckdb = pytest.importorskip("duckdb")


@frozen
class PinsResource:
    name = field()
    suffix = field(validator=optional(in_((".csv", ".parquet"))), default=None)

    def __attrs_post_init__(self):
        if self.suffix is None:
            object.__setattr__(self, "suffix", self.path.suffix)
        if self.path.suffix != self.suffix:
            raise ValueError(
                f"path suffix {self.path.suffix!r} does not match expected suffix {self.suffix!r}"
            )
        if self.name not in xo.options.pins.get_board().pin_list():
            raise ValueError(f"name {self.name!r} not found in pins board")

    @property
    def table_name(self):
        return f"test-{self.name}"

    @property
    @functools.cache
    def path(self):
        return pathlib.Path(xo.options.pins.get_path(self.name))

    def get_underlying_method(self, con):
        return getattr(con, self.deferred_reader.method_name)

    @property
    def deferred_reader(self):
        match self.suffix:
            case ".parquet":
                return deferred_read_parquet
            case ".csv":
                return deferred_read_csv
            case _:
                raise ValueError(f"unsupported suffix {self.suffix!r}")

    @property
    def immediate_reader(self):
        match self.suffix:
            case ".parquet":
                return pd.read_parquet
            case ".csv":
                return pd.read_csv
            case _:
                raise ValueError(f"unsupported suffix {self.suffix!r}")

    @property
    @functools.cache
    def df(self):
        return self.immediate_reader(self.path)


@pytest.fixture(scope="session")
def iris_csv():
    return PinsResource(name="iris", suffix=".csv")


@pytest.fixture(scope="session")
def astronauts_parquet():
    return PinsResource(name="astronauts", suffix=".parquet")


def filter_sepal_length(t):
    return t.sepal_length > 5


def filter_field21(t):
    return t.field21 > 2


def ensure_tmp_csv(csv_name, tmp_path):
    source_path = pathlib.Path(xo.options.pins.get_path(csv_name))
    target_path = tmp_path.joinpath(source_path.name)
    if not target_path.exists():
        target_path.write_text(source_path.read_text())
    return target_path


def mutate_csv(path, line=None):
    if line is None:
        line = path.read_text().strip().rsplit("\n", 1)[-1]
    with path.open("at") as fh:
        fh.writelines([line])


@pytest.mark.parametrize("pins_resource", ("iris_csv", "astronauts_parquet"))
def test_deferred_read_cache_key_check(con, tmp_path, pins_resource, request):
    # check that we don't invoke read when we calc key
    pins_resource = request.getfixturevalue(pins_resource)
    cache = ParquetCache.from_kwargs(source=xo.connect(), relative_path=tmp_path)

    assert pins_resource.table_name not in con.tables
    t = pins_resource.deferred_reader(pins_resource.path, con, pins_resource.table_name)
    cache.strategy.calc_key(t)
    assert pins_resource.table_name not in con.tables


@pytest.mark.parametrize("pins_resource", ("iris_csv", "astronauts_parquet"))
def test_deferred_read_to_sql(con, pins_resource, request):
    # check that we don't invoke read when we convert to sql
    pins_resource = request.getfixturevalue(pins_resource)
    assert pins_resource.table_name not in con.tables
    t = pins_resource.deferred_reader(pins_resource.path, con, pins_resource.table_name)
    xo.to_sql(t)
    assert pins_resource.table_name not in con.tables


@pytest.mark.parametrize(
    "get_con,pins_resource",
    [
        pytest.param(con, pins_resource, marks=marks)
        for con, marks in [
            (lambda: xo.pandas.connect(), []),
            (lambda: xo.postgres.connect_env(), [pytest.mark.postgres]),
        ]
        for pins_resource in ("iris_csv", "astronauts_parquet")
    ],
)
def test_deferred_read(get_con, pins_resource, request):
    from adbc_driver_manager import ProgrammingError

    con = get_con()
    pins_resource = request.getfixturevalue(pins_resource)
    assert pins_resource.table_name not in con.tables
    t = pins_resource.deferred_reader(pins_resource.path, con, pins_resource.table_name)
    assert xo.execute(t).equals(pins_resource.df)
    assert pins_resource.table_name in con.tables
    # is this a test of mode for postgres?
    if con.name != "pandas":
        # verify that we can't execute again (pandas happily clobbers)
        with pytest.raises(
            ProgrammingError,
            match=f'relation "{pins_resource.table_name}" already exists',
        ):
            assert xo.execute(t).equals(pins_resource.df)
    con.drop_table(pins_resource.table_name, force=True)
    assert pins_resource.table_name not in tuple(con.tables)


@pytest.mark.postgres
@pytest.mark.parametrize(
    "get_con,pins_resource",
    itertools.product(
        (lambda: xo.postgres.connect_env(),),
        ("iris_csv", "astronauts_parquet"),
    ),
)
def test_deferred_read_temporary(get_con, pins_resource, request):
    con = get_con()
    pins_resource = request.getfixturevalue(pins_resource)
    t = pins_resource.deferred_reader(pins_resource.path, con, None, temporary=True)
    table_name = t.op().name
    assert xo.execute(t).equals(pins_resource.df)
    assert table_name in con.tables
    con.drop_table(table_name)
    assert table_name not in con.tables


@pytest.mark.parametrize(
    "get_con,pins_resource,filter_",
    [
        pytest.param(con, pins_resource, filter_, marks=marks)
        for con, marks in [
            (lambda: xo.pandas.connect(), []),
            (lambda: xo.postgres.connect_env(), [pytest.mark.postgres]),
            (lambda: xo.duckdb.connect(), []),
        ]
        for (pins_resource, filter_) in (
            ("iris_csv", filter_sepal_length),
            ("astronauts_parquet", filter_field21),
        )
    ],
)
def test_cached_deferred_read(get_con, pins_resource, filter_, request, tmp_path):
    from adbc_driver_manager import ProgrammingError

    con = get_con()
    pins_resource = request.getfixturevalue(pins_resource)
    cache = ParquetCache.from_kwargs(source=xo.connect(), relative_path=tmp_path)

    df = pins_resource.df[filter_].reset_index(drop=True)
    t = pins_resource.deferred_reader(pins_resource.path, con, pins_resource.table_name)
    expr = t[filter_].cache(cache=cache)

    # no work is done yet
    assert pins_resource.table_name not in con.tables
    assert not cache.exists(expr)

    # something exists in both con and cache
    assert xo.execute(expr).equals(df)
    assert pins_resource.table_name in con.tables
    assert cache.exists(expr)

    # we read from cache even if the table disappears
    try:
        con.drop_table(t.op().name, force=True)
    except duckdb.duckdb.CatalogException:
        con.drop_view(t.op().name)

    assert xo.execute(expr).equals(df)
    assert pins_resource.table_name not in con.tables

    # we repopulate the cache
    cache.drop(expr)
    assert xo.execute(expr).equals(df)
    assert pins_resource.table_name in con.tables
    assert cache.exists(expr)

    if con.name == "postgres":
        # we are mode="create" by default, which means losing cache creates collision
        mode = get_partial_arguments(pins_resource.get_underlying_method(con))["mode"]
        assert mode == "create"
        cache.drop(expr)
        with pytest.raises(
            ProgrammingError,
            match=f'relation "{pins_resource.table_name}" already exists',
        ):
            xo.execute(expr)

        # with mode="replace" we can clobber
        t = pins_resource.deferred_reader(
            pins_resource.path, con, pins_resource.table_name, mode="replace"
        )
        expr = t[filter_].cache(cache=cache)
        assert xo.execute(expr).equals(df)
        assert cache.exists(expr)
        assert pins_resource.table_name in con.tables
        # this fails above, but works here because of mode="replace"
        cache.drop(expr)
        assert xo.execute(expr).equals(df)


@pytest.mark.parametrize(
    "get_con",
    [
        lambda: xo.pandas.connect(),
        pytest.param(lambda: xo.postgres.connect_env(), marks=pytest.mark.postgres),
    ],
)
def test_cached_csv_mutate(get_con, iris_csv, tmp_path):
    con = get_con()
    target_path = ensure_tmp_csv(iris_csv.name, tmp_path)
    cache = ParquetCache.from_kwargs(source=xo.connect(), relative_path=tmp_path)
    # make sure the con is "clean"
    if iris_csv.table_name in con.tables:
        con.drop_table(iris_csv.table_name, force=True)

    df = iris_csv.df
    kwargs = {"mode": "replace"} if con.name == "postgres" else {}
    t = iris_csv.deferred_reader(target_path, con, iris_csv.table_name, **kwargs)
    expr = t.cache(cache=cache)

    # nothing exists yet
    assert iris_csv.table_name not in con.tables
    assert not cache.exists(expr)

    # initial cache population
    assert xo.execute(expr).equals(df)
    assert iris_csv.table_name in con.tables
    assert cache.exists(expr)

    # mutate
    mutate_csv(target_path)
    df = iris_csv.immediate_reader(target_path)
    assert not cache.exists(expr)
    assert xo.execute(expr).equals(df)
    assert cache.exists(expr)


@pytest.mark.parametrize(
    "method_name,path",
    [
        (
            "deferred_read_csv",
            "https://raw.githubusercontent.com/ibis-project/testing-data/refs/heads/master/csv/astronauts.csv",
        ),
        (
            "deferred_read_parquet",
            "https://nasa-avionics-data-ml.s3.us-east-2.amazonaws.com/Tail_652_1_parquet/652200101120916.16p0.parquet",
        ),
    ],
)
@pytest.mark.parametrize(
    "remote",
    [True, False],
)
def test_deferred_read_cache(con, tmp_path, method_name, path, remote):
    cache = ParquetCache.from_kwargs(source=xo.connect(), relative_path=tmp_path)
    read_method = getattr(xo, method_name)
    connection = con if remote else xo.duckdb.connect()

    t = read_method(path, connection)
    uncached = t.head(10)
    assert cache.strategy.calc_key(uncached) is not None

    expr = uncached.cache(cache=cache)
    assert not expr.execute().empty


@pytest.mark.postgres
def test_deferred_read_kwargs(pg):
    name = "iris"
    read0, read1 = (
        xo.examples.get_table_from_name(name, pg, mode=mode)
        for mode in ("create", "replace")
    )
    hash0, hash1 = (dask.base.tokenize(expr) for expr in (read0, read1))
    assert hash0 != hash1


def test_deferred_read_parquet_multiple_paths(parquet_dir):
    path = str(parquet_dir / "astronauts.parquet")
    expr = deferred_read_parquet((path, path), xo.connect())
    assert not expr.execute().empty


def test_deferred_read_csv_multiple_paths(csv_dir):
    path = str(csv_dir / "astronauts.csv")
    con = xo.connect()

    t = con.read_csv(path)

    expr = deferred_read_csv((path, path), con, schema=t.schema())

    assert not expr.execute().empty


@pytest.fixture(scope="function")
def backend(request, con):
    lookup = {"duckdb": xo.duckdb.connect(), "postgres": con, "xorq": xo.connect()}

    return lookup.get(request.param, con)


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("duckdb", id="duckdb"),
        pytest.param("postgres", id="postgres"),
        pytest.param("xorq", id="xorq"),
    ],
    indirect=True,
)
def test_register_csv_with_glob_string(data_dir, backend):
    table_name = f"{backend.name}_astronauts"
    glob_pattern = str(data_dir / "csv" / "*astronauts.csv")
    expected = backend.read_csv(
        glob_pattern, table_name=f"{table_name}_expected"
    ).execute()

    read = xo.deferred_read_csv(glob_pattern, backend, table_name=table_name)
    actual = read.execute()  # triggers the table creation

    assert any(table_name in t for t in backend.list_tables())
    assert_frame_equal(expected, actual)


def test_register_empty_glob_pattern_fails(data_dir, con):
    glob_pattern = str(data_dir / "csv" / "*foo.csv")

    with pytest.raises(ValueError, match="At least one path is required"):
        xo.deferred_read_csv(glob_pattern, con, table_name="foo")



try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata


def _load_entry_points():
    eps = importlib_metadata.entry_points(group="xorq.backends")
    return sorted(eps)


def load_backend(name):
    if entry_point := next(
        (ep for ep in _load_entry_points() if ep.name == name), None
    ):
        import types

        module = entry_point.load()
        backend = module.Backend()
        backend.register_options()

        def connect(*args, **kwargs):
            return backend.connect(*args, **kwargs)

        connect.__doc__ = backend.do_connect.__doc__
        connect.__wrapped__ = backend.do_connect
        connect.__module__ = f"xorq.{name}"

        proxy = types.ModuleType(f"xorq.{name}")
        proxy.connect = connect
        proxy.compile = backend.compile
        proxy.has_operation = backend.has_operation
        proxy.name = name
        proxy._from_url = backend._from_url

        # Add any additional methods that should be exposed at the top level
        for attr in getattr(backend, "_top_level_methods", ()):
            setattr(proxy, attr, getattr(backend, attr))

        return proxy



import pandas as pd
import pyarrow as pa
from datafusion import udwf
from datafusion.user_defined import WindowEvaluator

import xorq.api as xo
from xorq.expr.relations import into_backend


class SmoothTwoColumn(WindowEvaluator):
    """Smooth once column based on a condition of another column.

    If the second column is above a threshold, then smooth over the first column from
    the previous and next rows.
    """

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def evaluate_all(self, values: list[pa.Array], num_rows: int) -> pa.Array:
        results = []
        values_a = values[0]
        values_b = values[1]
        for idx in range(num_rows):
            if not values_b[idx].is_valid:
                if idx == 0:
                    results.append(values_a[1].cast(pa.float64()))
                elif idx == num_rows - 1:
                    results.append(values_a[num_rows - 2].cast(pa.float64()))
                else:
                    results.append(
                        pa.scalar(
                            values_a[idx - 1].as_py() * self.alpha
                            + values_a[idx + 1].as_py() * (1.0 - self.alpha)
                        )
                    )
            else:
                results.append(values_a[idx].cast(pa.float64()))

        return pa.array(results)


def test_cross_engine_analytics(con, ddb_con):
    # register udwf in bare datafusion
    context = con.con
    smooth_two_col = udwf(
        lambda: SmoothTwoColumn(0.9),
        [pa.float64(), pa.int64()],
        pa.float64(),
        volatility="immutable",
        name="smooth_two_col",
    )
    context.register_udwf(smooth_two_col)

    # create new table with the new computed column
    query = "select trades.*, smooth_two_col(trades.price, trades.volume) over (partition by trades.symbol) as vwap from trades"
    vwap_df = con.raw_sql(query).to_pandas()
    trades = con.create_table("vwap", vwap_df, temp=False).pipe(into_backend, ddb_con)

    # Perform asof join in DuckDB
    asof_result = (
        ddb_con.table("quotes")
        .asof_join(trades, predicates=["symbol"], on="timestamp")
        .pipe(xo.execute)
    )

    assert isinstance(asof_result, pd.DataFrame)
    assert len(asof_result) > 0



"""Reduction operations."""

from __future__ import annotations

from typing import Literal, Optional

from public import public

import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.core import Column, Value
from xorq.vendor.ibis.expr.operations.relations import Relation  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.sortkeys import SortKey  # noqa: TCH001


@public
class Reduction(Value):
    """Base class for reduction operations."""

    shape = ds.scalar


# TODO(kszucs): all reductions all filterable so we could remove Filterable
class Filterable(Value):
    where: Optional[Value[dt.Boolean]] = None


@public
class Count(Filterable, Reduction):
    """Count the number of non-null elements of a column."""

    arg: Column[dt.Any]

    dtype = dt.int64


@public
class CountStar(Filterable, Reduction):
    """Count the number of rows of a relation."""

    arg: Relation

    dtype = dt.int64

    @attribute
    def relations(self):
        return frozenset({self.arg})


@public
class CountDistinctStar(Filterable, Reduction):
    """Count the number of distinct rows of a relation."""

    arg: Relation

    dtype = dt.int64

    @attribute
    def relations(self):
        return frozenset({self.arg})


@public
class Arbitrary(Filterable, Reduction):
    """Retrieve an arbitrary element.

    Returns a non-null value unless the column is empty or all values are NULL.
    """

    arg: Column[dt.Any]

    dtype = rlz.dtype_like("arg")


@public
class First(Filterable, Reduction):
    """Retrieve the first element."""

    arg: Column[dt.Any]
    order_by: VarTuple[SortKey] = ()
    include_null: bool = False

    dtype = rlz.dtype_like("arg")


@public
class Last(Filterable, Reduction):
    """Retrieve the last element."""

    arg: Column[dt.Any]
    order_by: VarTuple[SortKey] = ()
    include_null: bool = False

    dtype = rlz.dtype_like("arg")


@public
class BitAnd(Filterable, Reduction):
    """Aggregate bitwise AND operation.

    All elements in an integer column are ANDed together.

    This can be used to determine which bit flags are set on all elements.

    See Also
    --------
    * BigQuery [`BIT_AND`](https://cloud.google.com/bigquery/docs/reference/standard-sql/aggregate_functions#bit_and)
    * MySQL [`BIT_AND`](https://dev.mysql.com/doc/refman/5.7/en/aggregate-functions.html#function_bit-and)
    """

    arg: Column[dt.Integer]

    dtype = rlz.dtype_like("arg")


@public
class BitOr(Filterable, Reduction):
    """Aggregate bitwise OR operation.

    All elements in an integer column are ORed together. This can be used
    to determine which bit flags are set on any element.

    See Also
    --------
    * BigQuery [`BIT_OR`](https://cloud.google.com/bigquery/docs/reference/standard-sql/aggregate_functions#bit_or)
    * MySQL [`BIT_OR`](https://dev.mysql.com/doc/refman/5.7/en/aggregate-functions.html#function_bit-or)
    """

    arg: Column[dt.Integer]

    dtype = rlz.dtype_like("arg")


@public
class BitXor(Filterable, Reduction):
    """Aggregate bitwise XOR operation.

    All elements in an integer column are XORed together. This can be used
    as a parity checksum of element values.

    See Also
    --------
    * BigQuery [`BIT_XOR`](https://cloud.google.com/bigquery/docs/reference/standard-sql/aggregate_functions#bit_xor)
    * MySQL [`BIT_XOR`](https://dev.mysql.com/doc/refman/5.7/en/aggregate-functions.html#function_bit-xor)
    """

    arg: Column[dt.Integer]

    dtype = rlz.dtype_like("arg")


@public
class Sum(Filterable, Reduction):
    """Compute the sum of a column."""

    arg: Column[dt.Numeric | dt.Boolean]

    @attribute
    def dtype(self):
        dtype = self.arg.dtype
        if dtype.is_boolean():
            return dt.int64
        elif dtype.is_integer():
            return dt.int64
        elif dtype.is_unsigned_integer():
            return dt.uint64
        elif dtype.is_floating():
            return dt.float64
        elif dtype.is_decimal():
            return dt.Decimal(
                precision=max(dtype.precision, 38)
                if dtype.precision is not None
                else None,
                scale=max(dtype.scale, 2) if dtype.scale is not None else None,
            )
        else:
            raise TypeError(f"Cannot compute sum of {dtype} values")


@public
class Mean(Filterable, Reduction):
    """Compute the mean of a column."""

    arg: Column[dt.Numeric | dt.Boolean]

    @attribute
    def dtype(self):
        if (dtype := self.arg.dtype).is_boolean():
            return dt.float64
        else:
            return dt.higher_precedence(dtype, dt.float64)


class QuantileBase(Filterable, Reduction):
    arg: Column

    @attribute
    def dtype(self):
        dtype = self.arg.dtype
        if dtype.is_numeric():
            dtype = dt.higher_precedence(dtype, dt.float64)
        return dtype


@public
class Median(QuantileBase):
    """Compute the median of a column."""


@public
class ApproxMedian(Median):
    """Compute the approximate median of a column."""


@public
class Quantile(QuantileBase):
    """Compute the quantile of a column."""

    quantile: Value[dt.Numeric]


@public
class ApproxQuantile(Quantile):
    """Compute the approximate quantile of a column."""

    arg: Column[dt.Numeric]


@public
class MultiQuantile(Filterable, Reduction):
    """Compute multiple quantiles of a column."""

    arg: Column
    quantile: Value[dt.Array[dt.Numeric]]

    @attribute
    def dtype(self):
        dtype = self.arg.dtype
        if dtype.is_numeric():
            dtype = dt.higher_precedence(dtype, dt.float64)
        return dt.Array(dtype)


@public
class ApproxMultiQuantile(MultiQuantile):
    """Compute multiple approximate quantiles of a column."""

    arg: Column[dt.Numeric]


class VarianceBase(Filterable, Reduction):
    """Base class for variance and standard deviation."""

    arg: Column[dt.Numeric | dt.Boolean]
    how: Literal["sample", "pop"]

    @attribute
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.float64


@public
class StandardDev(VarianceBase):
    """Compute the standard deviation of a column."""


@public
class Variance(VarianceBase):
    """Compute the variance of a column."""


@public
class Correlation(Filterable, Reduction):
    """Correlation coefficient of two columns."""

    left: Column[dt.Numeric | dt.Boolean]
    right: Column[dt.Numeric | dt.Boolean]
    how: Literal["sample", "pop"] = "sample"

    dtype = dt.float64


@public
class Covariance(Filterable, Reduction):
    """Covariance of two columns."""

    left: Column[dt.Numeric | dt.Boolean]
    right: Column[dt.Numeric | dt.Boolean]
    how: Literal["sample", "pop"]

    dtype = dt.float64


@public
class Mode(Filterable, Reduction):
    """Compute the mode of a column."""

    arg: Column

    dtype = rlz.dtype_like("arg")


@public
class Max(Filterable, Reduction):
    """Compute the maximum of a column."""

    arg: Column

    dtype = rlz.dtype_like("arg")


@public
class Min(Filterable, Reduction):
    """Compute the minimum of a column."""

    arg: Column

    dtype = rlz.dtype_like("arg")


@public
class ArgMax(Filterable, Reduction):
    """Compute the index of the maximum value in a column."""

    arg: Column
    key: Column

    dtype = rlz.dtype_like("arg")


@public
class ArgMin(Filterable, Reduction):
    """Compute the index of the minimum value in a column."""

    arg: Column
    key: Column

    dtype = rlz.dtype_like("arg")


@public
class GroupConcat(Filterable, Reduction):
    """Concatenate strings in a group with a given separator character."""

    arg: Column
    sep: Value[dt.String]
    order_by: VarTuple[SortKey] = ()

    dtype = dt.string


@public
class CountDistinct(Filterable, Reduction):
    """Count the number of distinct values in a column."""

    arg: Column

    dtype = dt.int64


@public
class ApproxCountDistinct(CountDistinct):
    """Approximate number of unique values."""


@public
class ArrayCollect(Filterable, Reduction):
    """Collect values into an array."""

    arg: Column
    order_by: VarTuple[SortKey] = ()
    include_null: bool = False

    @attribute
    def dtype(self):
        return dt.Array(self.arg.dtype)


@public
class All(Filterable, Reduction):
    """Check if all values in a column are true."""

    arg: Column[dt.Boolean]

    dtype = dt.boolean


@public
class Any(Filterable, Reduction):
    """Check if any value in a column is true."""

    arg: Column[dt.Boolean]

    dtype = dt.boolean



"""Tests for deferred sklearn metrics evaluation using Pipeline API."""

import numpy as np
import pandas as pd
import pytest

import xorq.expr.datatypes as dt
from xorq.expr import api
from xorq.expr.ml.enums import ResponseMethod
from xorq.expr.ml.metrics import deferred_sklearn_metric
from xorq.expr.ml.pipeline_lib import Pipeline


# Skip all tests in this module if sklearn is not installed
sklearn = pytest.importorskip("sklearn")

# Access sklearn modules through the sklearn object to avoid E402 linting errors
make_classification = sklearn.datasets.make_classification
make_regression = sklearn.datasets.make_regression
RandomForestClassifier = sklearn.ensemble.RandomForestClassifier
RandomForestRegressor = sklearn.ensemble.RandomForestRegressor
LinearRegression = sklearn.linear_model.LinearRegression
LogisticRegression = sklearn.linear_model.LogisticRegression
LinearSVC = sklearn.svm.LinearSVC
StandardScaler = sklearn.preprocessing.StandardScaler
SkPipeline = sklearn.pipeline.Pipeline
accuracy_score = sklearn.metrics.accuracy_score
mean_absolute_error = sklearn.metrics.mean_absolute_error
mean_squared_error = sklearn.metrics.mean_squared_error
precision_score = sklearn.metrics.precision_score
r2_score = sklearn.metrics.r2_score
recall_score = sklearn.metrics.recall_score
roc_auc_score = sklearn.metrics.roc_auc_score
f1_score = sklearn.metrics.f1_score
train_test_split = sklearn.model_selection.train_test_split


@pytest.fixture
def classification_data():
    """Generate synthetic classification data."""
    X, y = make_classification(
        n_samples=1000,
        n_features=10,
        n_informative=5,
        n_redundant=3,
        n_classes=2,
        random_state=42,
    )

    feature_names = [f"feature_{i}" for i in range(X.shape[1])]
    df = pd.DataFrame(X, columns=feature_names)
    df["target"] = y

    train_df, test_df = train_test_split(df, test_size=0.3, random_state=42)

    return train_df, test_df, feature_names


@pytest.fixture
def multiclass_data():
    """Generate synthetic multiclass data."""
    X, y = make_classification(
        n_samples=1000,
        n_features=10,
        n_informative=5,
        n_redundant=3,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=42,
    )

    feature_names = [f"feature_{i}" for i in range(X.shape[1])]
    df = pd.DataFrame(X, columns=feature_names)
    df["target"] = y

    train_df, test_df = train_test_split(df, test_size=0.3, random_state=42)

    return train_df, test_df, feature_names


@pytest.fixture
def regression_data():
    """Generate synthetic regression data."""
    X, y = make_regression(
        n_samples=1000, n_features=10, n_informative=5, noise=10.0, random_state=42
    )

    feature_names = [f"feature_{i}" for i in range(X.shape[1])]
    df = pd.DataFrame(X, columns=feature_names)
    df["target"] = y

    train_df, test_df = train_test_split(df, test_size=0.3, random_state=42)

    return train_df, test_df, feature_names


def test_pipeline_classification_metrics(classification_data):
    """Test Pipeline API with classification metrics."""
    train_df, test_df, feature_names = classification_data

    # Register data with xorq
    train_expr = api.register(train_df, "train")
    test_expr = api.register(test_df, "test")

    # Create sklearn pipeline with scaler and classifier
    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("classifier", RandomForestClassifier(n_estimators=10, random_state=42)),
        ]
    )

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        train_expr, features=feature_names, target="target"
    )

    # Direct sklearn metrics for comparison
    X_train = train_df[feature_names]
    y_train = train_df["target"]
    X_test = test_df[feature_names]
    y_test = test_df["target"]

    sklearn_pipeline.fit(X_train, y_train)
    y_pred = sklearn_pipeline.predict(X_test)

    expected_accuracy = accuracy_score(y_test, y_pred)
    expected_precision = precision_score(y_test, y_pred)
    expected_recall = recall_score(y_test, y_pred)
    expected_f1 = f1_score(y_test, y_pred)

    # Get predictions once for all metrics
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Test accuracy using Pipeline API
    accuracy_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=accuracy_score,
    )
    actual_accuracy = accuracy_expr.execute()
    assert np.isclose(actual_accuracy, expected_accuracy)

    # Test precision
    precision_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=precision_score,
    )
    actual_precision = precision_expr.execute()
    assert np.isclose(actual_precision, expected_precision)

    # Test recall
    recall_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=recall_score,
    )
    actual_recall = recall_expr.execute()
    assert np.isclose(actual_recall, expected_recall)

    # Test F1 score
    f1_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=f1_score,
    )
    actual_f1 = f1_expr.execute()
    assert np.isclose(actual_f1, expected_f1)


def test_pipeline_regression_metrics(regression_data):
    """Test Pipeline API with regression metrics."""
    train_df, test_df, feature_names = regression_data

    # Register data with xorq
    train_expr = api.register(train_df, "train")
    test_expr = api.register(test_df, "test")

    # Create sklearn pipeline with scaler and regressor
    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("regressor", RandomForestRegressor(n_estimators=10, random_state=42)),
        ]
    )

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        train_expr, features=feature_names, target="target"
    )

    # Direct sklearn metrics for comparison
    X_train = train_df[feature_names]
    y_train = train_df["target"]
    X_test = test_df[feature_names]
    y_test = test_df["target"]

    sklearn_pipeline.fit(X_train, y_train)
    y_pred = sklearn_pipeline.predict(X_test)

    expected_mse = mean_squared_error(y_test, y_pred)
    expected_mae = mean_absolute_error(y_test, y_pred)
    expected_r2 = r2_score(y_test, y_pred)

    # Get predictions once for all metrics
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Test MSE using Pipeline API
    mse_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=mean_squared_error,
    )
    actual_mse = mse_expr.execute()
    assert np.isclose(actual_mse, expected_mse, rtol=1e-4)

    # Test MAE using Pipeline API
    mae_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=mean_absolute_error,
    )
    actual_mae = mae_expr.execute()
    assert np.isclose(actual_mae, expected_mae, rtol=1e-4)

    # Test R2 using Pipeline API
    r2_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=r2_score,
    )
    actual_r2 = r2_expr.execute()
    assert np.isclose(actual_r2, expected_r2, rtol=1e-4)


def test_pipeline_multiclass_metrics(multiclass_data):
    """Test Pipeline API with multiclass metrics."""
    train_df, test_df, feature_names = multiclass_data

    # Register data with xorq
    train_expr = api.register(train_df, "train")
    test_expr = api.register(test_df, "test")

    # Create sklearn pipeline with scaler and classifier
    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("classifier", LogisticRegression(max_iter=1000, random_state=42)),
        ]
    )

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        train_expr, features=feature_names, target="target"
    )

    # Direct sklearn metrics for comparison
    X_train = train_df[feature_names]
    y_train = train_df["target"]
    X_test = test_df[feature_names]
    y_test = test_df["target"]

    sklearn_pipeline.fit(X_train, y_train)
    y_pred = sklearn_pipeline.predict(X_test)

    expected_accuracy = accuracy_score(y_test, y_pred)
    expected_precision_macro = precision_score(
        y_test, y_pred, average="macro", zero_division=0
    )
    expected_recall_weighted = recall_score(
        y_test, y_pred, average="weighted", zero_division=0
    )

    # Get predictions once for all metrics
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Test accuracy
    accuracy_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=accuracy_score,
    )
    actual_accuracy = accuracy_expr.execute()
    assert np.isclose(actual_accuracy, expected_accuracy)

    # Test precision with macro averaging
    precision_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=precision_score,
        metric_kwargs={"average": "macro", "zero_division": 0},
    )
    actual_precision = precision_expr.execute()
    assert np.isclose(actual_precision, expected_precision_macro)

    # Test recall with weighted averaging
    recall_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=recall_score,
        metric_kwargs={"average": "weighted", "zero_division": 0},
    )
    actual_recall = recall_expr.execute()
    assert np.isclose(actual_recall, expected_recall_weighted)


def test_custom_metric_with_pipeline():
    """Test Pipeline API with a custom metric function."""
    # Create simple test data
    df = pd.DataFrame(
        {
            "feature_0": [1, 2, 3, 4, 5],
            "feature_1": [2, 3, 4, 5, 6],
            "target": [0, 0, 1, 1, 1],
        }
    )

    test_expr = api.register(df, "test")
    feature_names = ["feature_0", "feature_1"]

    # Create sklearn pipeline
    sklearn_pipeline = SkPipeline([("classifier", LogisticRegression(random_state=42))])

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        test_expr, features=feature_names, target="target"
    )

    # Define a custom metric wrapped with make_scorer
    from sklearn.metrics import make_scorer

    def custom_metric(y_true, y_pred):
        """A custom metric that returns accuracy * 100."""
        return accuracy_score(y_true, y_pred) * 100

    custom_scorer = make_scorer(custom_metric)

    # Get predictions
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Test custom metric via make_scorer
    custom_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=custom_scorer,
        return_type=dt.float64,
    )
    result = custom_expr.execute()

    # Should return a float between 0 and 100
    assert isinstance(result, (float, np.floating))
    assert 0 <= result <= 100


def test_deferred_nature_with_pipeline():
    """Test that the Pipeline API maintains deferred execution."""
    # Create simple test data
    df = pd.DataFrame(
        {
            "feature_0": range(100),
            "feature_1": range(100),
            "target": [i % 2 for i in range(100)],
        }
    )

    test_expr = api.register(df, "test")
    feature_names = ["feature_0", "feature_1"]

    # Create sklearn pipeline
    sklearn_pipeline = SkPipeline([("classifier", LogisticRegression(random_state=42))])

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        test_expr, features=feature_names, target="target"
    )

    # Get predictions
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Create a deferred metric expression
    accuracy_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=accuracy_score,
    )

    # The expression should have an execute method (deferred)
    assert hasattr(accuracy_expr, "execute")

    # It should not be a computed value yet
    assert not isinstance(accuracy_expr, (float, int, np.number))

    # Can create multiple metrics without execution
    precision_expr = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=precision_score,
        metric_kwargs={"average": "macro", "zero_division": 0},
    )

    # Both should be expressions
    assert hasattr(precision_expr, "execute")

    # Now execute one
    accuracy_value = accuracy_expr.execute()
    assert isinstance(accuracy_value, (float, np.floating))

    # The other is still deferred
    assert hasattr(precision_expr, "execute")
    assert not isinstance(precision_expr, (float, int, np.number))


def test_predict_proba_metrics():
    """Test that probability-based metrics work with predict_proba."""
    # Create binary classification data
    from sklearn.datasets import make_classification

    X, y = make_classification(
        n_samples=200, n_features=5, n_informative=3, n_classes=2, random_state=42
    )

    df = pd.DataFrame(X, columns=[f"feature_{i}" for i in range(X.shape[1])])
    df["target"] = y

    train_df = df.iloc[:150]
    test_df = df.iloc[150:]

    train_expr = api.register(train_df, "train")
    test_expr = api.register(test_df, "test")

    # Create pipeline with classifier that supports predict_proba
    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("classifier", LogisticRegression(random_state=42)),
        ]
    )

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)
    features = [f"feature_{i}" for i in range(X.shape[1])]

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(train_expr, features=features, target="target")

    # Get sklearn predictions for comparison
    X_train = train_df[features].values
    y_train = train_df["target"].values
    X_test = test_df[features].values
    y_test = test_df["target"].values

    sklearn_pipeline.fit(X_train, y_train)
    y_proba = sklearn_pipeline.predict_proba(X_test)[:, 1]

    expected_auc = roc_auc_score(y_test, y_proba)

    # Test AUC with predict_proba - explicit API
    expr_with_proba = fitted_pipeline.predict_proba(test_expr)
    auc_expr = deferred_sklearn_metric(
        expr=expr_with_proba,
        target="target",
        pred=ResponseMethod.PREDICT_PROBA,
        metric=roc_auc_score,
    )

    actual_auc = auc_expr.execute()

    # Verify AUC matches
    assert np.isclose(actual_auc, expected_auc, rtol=1e-5), (
        f"AUC mismatch: {actual_auc} vs {expected_auc}"
    )

    # Also test that we can get the predict_proba expression directly
    proba_expr = fitted_pipeline.predict_proba(test_expr)
    assert hasattr(proba_expr, "execute")

    # Execute and check shape
    proba_result = proba_expr[ResponseMethod.PREDICT_PROBA].execute()
    assert len(proba_result) == len(test_df)


def test_probability_metric_without_predict_proba():
    X, y = make_classification(
        n_samples=200, n_features=5, n_informative=3, n_classes=2, random_state=7
    )
    df = pd.DataFrame(X, columns=[f"feature_{i}" for i in range(X.shape[1])])
    df["target"] = y

    train_df = df.iloc[:150]
    test_df = df.iloc[150:]

    train_expr = api.register(train_df, "svm_train")
    test_expr = api.register(test_df, "svm_test")

    features = [f"feature_{i}" for i in range(X.shape[1])]

    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("svm", LinearSVC(random_state=7, max_iter=5000)),
        ]
    )

    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)
    fitted_pipeline = xorq_pipeline.fit(train_expr, features=features, target="target")

    X_train = train_df[features].values
    y_train = train_df["target"].values
    X_test = test_df[features].values
    y_test = test_df["target"].values

    sklearn_pipeline.fit(X_train, y_train)
    y_scores = sklearn_pipeline.decision_function(X_test)

    # ROC-AUC works with raw decision function scores for binary classification
    expected_auc = roc_auc_score(y_test, y_scores)

    # Use decision_function - works directly with ROC-AUC for binary
    expr_with_scores = fitted_pipeline.decision_function(test_expr)
    auc_expr = deferred_sklearn_metric(
        expr=expr_with_scores,
        target="target",
        pred=ResponseMethod.DECISION_FUNCTION,
        metric=roc_auc_score,
    )

    actual_auc = auc_expr.execute()
    assert np.isclose(actual_auc, expected_auc)


def test_probability_metric_multiclass_with_predict_proba():
    """Test multi-class ROC-AUC using a classifier with predict_proba.

    For multi-class ROC-AUC, sklearn requires probabilities that sum to 1.
    We use LogisticRegression which natively supports predict_proba.
    """
    X, y = make_classification(
        n_samples=240,
        n_features=6,
        n_informative=4,
        n_redundant=0,
        n_classes=3,
        random_state=11,
    )
    df = pd.DataFrame(X, columns=[f"feature_{i}" for i in range(X.shape[1])])
    df["target"] = y

    train_df = df.iloc[:180]
    test_df = df.iloc[180:]

    train_expr = api.register(train_df, "multiclass_train")
    test_expr = api.register(test_df, "multiclass_test")

    features = [f"feature_{i}" for i in range(X.shape[1])]

    # Use LogisticRegression which has predict_proba
    sklearn_pipeline = SkPipeline(
        [
            ("scaler", StandardScaler()),
            ("clf", LogisticRegression(random_state=11, max_iter=5000)),
        ]
    )

    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)
    fitted_pipeline = xorq_pipeline.fit(train_expr, features=features, target="target")

    X_train = train_df[features].values
    y_train = train_df["target"].values
    X_test = test_df[features].values
    y_test = test_df["target"].values

    sklearn_pipeline.fit(X_train, y_train)
    sklearn_proba = sklearn_pipeline.predict_proba(X_test)

    expected_auc = roc_auc_score(y_test, sklearn_proba, multi_class="ovr")

    # Use predict_proba - sklearn handles normalization internally
    expr_with_proba = fitted_pipeline.predict_proba(test_expr)
    auc_expr = deferred_sklearn_metric(
        expr=expr_with_proba,
        target="target",
        pred=ResponseMethod.PREDICT_PROBA,
        metric=roc_auc_score,
        metric_kwargs={"multi_class": "ovr"},
    )

    actual_auc = auc_expr.execute()
    assert np.isclose(actual_auc, expected_auc)


def test_metric_kwargs_with_pipeline():
    """Test that metric kwargs work correctly with Pipeline API."""
    # Create simple test data
    df = pd.DataFrame(
        {
            "feature_0": range(20),
            "feature_1": range(20, 40),
            "target": [i % 3 for i in range(20)],  # 3 classes
        }
    )

    test_expr = api.register(df, "test")
    feature_names = ["feature_0", "feature_1"]

    # Create sklearn pipeline
    sklearn_pipeline = SkPipeline(
        [("classifier", LogisticRegression(random_state=42, max_iter=200))]
    )

    # Wrap with xorq Pipeline
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)

    # Fit the pipeline
    fitted_pipeline = xorq_pipeline.fit(
        test_expr, features=feature_names, target="target"
    )

    # Get predictions
    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Test precision with different averaging methods
    precision_macro = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=precision_score,
        metric_kwargs={"average": "macro", "zero_division": 0},
    ).execute()

    precision_weighted = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=precision_score,
        metric_kwargs={"average": "weighted", "zero_division": 0},
    ).execute()

    # Different averaging should give different results
    assert precision_macro != precision_weighted

    # Both should be valid precision scores
    assert 0 <= precision_macro <= 1
    assert 0 <= precision_weighted <= 1


def test_sign_auto_detected():
    """Test that sign is automatically extracted from the scorer.

    Bare callable (mean_squared_error) resolves with sign=1.
    String scorer ("neg_mean_squared_error") resolves with sign=-1.
    """
    df = pd.DataFrame(
        {
            "feature_0": [0.0, 1.0, 2.0, 3.0],
            "feature_1": [0.0, 1.0, 2.0, 3.0],
            "target": [0.0, 1.0, 2.0, 3.0],
        }
    )

    test_expr = api.register(df, "sign_test")
    feature_names = ["feature_0", "feature_1"]

    sklearn_pipeline = SkPipeline([("regressor", LinearRegression())])
    xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)
    fitted_pipeline = xorq_pipeline.fit(
        test_expr, features=feature_names, target="target"
    )

    expr_with_preds = fitted_pipeline.predict(test_expr)

    # Bare callable — sign=1 (raw metric value)
    mse_positive = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric=mean_squared_error,
    ).execute()

    # String scorer — sign=-1 (negated by convention)
    mse_negative = deferred_sklearn_metric(
        expr=expr_with_preds,
        target="target",
        pred=ResponseMethod.PREDICT,
        metric="neg_mean_squared_error",
    ).execute()

    assert mse_negative == -mse_positive
    assert mse_positive >= 0  # MSE is always non-negative


def test_scorer_from_spec_string_input():
    """str -> resolved scorer with sign, kwargs, response_method."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("accuracy")
    assert s.metric_fn is accuracy_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_neg_scorer():
    """neg_* string -> sign=-1."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("neg_mean_squared_error")
    assert s.metric_fn is mean_squared_error
    assert s.sign == -1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_proba_scorer():
    """roc_auc string -> response_method from scorer."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("roc_auc")
    assert s.metric_fn is roc_auc_score
    assert s.sign == 1
    # roc_auc uses decision_function or predict_proba
    assert s.response_method in (
        ResponseMethod.DECISION_FUNCTION,
        ResponseMethod.PREDICT_PROBA,
    )


def test_scorer_from_spec_base_scorer_input():
    """_BaseScorer -> extracts metric_fn, sign, kwargs."""
    from sklearn.metrics import make_scorer

    from xorq.expr.ml.metrics import Scorer

    scorer_obj = make_scorer(accuracy_score)
    s = Scorer.from_spec(scorer_obj)
    assert s.metric_fn is accuracy_score
    assert s.sign == 1


def test_scorer_from_spec_known_callable():
    """Known bare callable -> sign=1, kwargs=(), response_method=predict."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec(accuracy_score)
    assert s.metric_fn is accuracy_score
    assert s.sign == 1
    assert s.kwargs == ()
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_unknown_callable_raises():
    """Unknown callable -> ValueError."""
    from sklearn.metrics import confusion_matrix

    from xorq.expr.ml.metrics import Scorer

    with pytest.raises(ValueError, match="not a known sklearn scorer function"):
        Scorer.from_spec(confusion_matrix)


def test_scorer_from_spec_scorer_passthrough():
    """Scorer input -> returns same instance."""
    from xorq.expr.ml.metrics import Scorer

    original = Scorer(
        metric_fn=accuracy_score,
        sign=1,
        kwargs={},
        response_method=ResponseMethod.PREDICT,
    )
    result = Scorer.from_spec(original)
    assert result is original


def test_scorer_from_model_classifier():
    """Classifier -> accuracy_score."""
    from sklearn.linear_model import LogisticRegression

    from xorq.expr.ml.metrics import Scorer

    model = LogisticRegression()
    model.fit([[0], [1]], [0, 1])
    s = Scorer.from_model(model)
    assert s.metric_fn is accuracy_score


def test_scorer_from_model_regressor():
    """Regressor -> r2_score."""
    from sklearn.linear_model import LinearRegression

    from xorq.expr.ml.metrics import Scorer

    model = LinearRegression()
    model.fit([[0], [1]], [0.0, 1.0])
    s = Scorer.from_model(model)
    assert s.metric_fn is r2_score


def test_scorer_from_model_cluster():
    """Clusterer -> adjusted_rand_score."""
    from sklearn.cluster import KMeans
    from sklearn.metrics import adjusted_rand_score

    from xorq.expr.ml.metrics import Scorer

    model = KMeans(n_clusters=2, n_init=1)
    model.fit([[0], [1]])
    s = Scorer.from_model(model)
    assert s.metric_fn is adjusted_rand_score


def test_scorer_from_spec_invalid_input_raises():
    """Non-callable, non-string -> ValueError."""
    from xorq.expr.ml.metrics import Scorer

    with pytest.raises(ValueError, match="scorer must be"):
        Scorer.from_spec(42)


def test_scorer_from_spec_string_neg_mean_absolute_error():
    """neg_mean_absolute_error string -> sign=-1, metric_fn=mean_absolute_error."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("neg_mean_absolute_error")
    assert s.metric_fn is mean_absolute_error
    assert s.sign == -1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_r2():
    """r2 string -> sign=1, metric_fn=r2_score."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("r2")
    assert s.metric_fn is r2_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_precision():
    """precision string -> sign=1, metric_fn=precision_score."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("precision")
    assert s.metric_fn is precision_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_recall():
    """recall string -> sign=1, metric_fn=recall_score."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("recall")
    assert s.metric_fn is recall_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_string_f1():
    """f1 string -> sign=1, metric_fn=f1_score."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec("f1")
    assert s.metric_fn is f1_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_callable_mean_squared_error():
    """Known bare callable mean_squared_error -> sign=1."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec(mean_squared_error)
    assert s.metric_fn is mean_squared_error
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_callable_r2_score():
    """Known bare callable r2_score -> sign=1."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec(r2_score)
    assert s.metric_fn is r2_score
    assert s.sign == 1
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_callable_roc_auc_score():
    """Known bare callable roc_auc_score -> sign=1, response_method=predict."""
    from xorq.expr.ml.metrics import Scorer

    s = Scorer.from_spec(roc_auc_score)
    assert s.metric_fn is roc_auc_score
    assert s.sign == 1
    # bare callable doesn't carry response_method metadata
    assert s.response_method == ResponseMethod.PREDICT


def test_scorer_from_spec_make_scorer_with_kwargs():
    """make_scorer with extra kwargs -> kwargs preserved."""
    from sklearn.metrics import make_scorer

    from xorq.expr.ml.metrics import Scorer

    scorer_obj = make_scorer(precision_score, average="macro", zero_division=0)
    s = Scorer.from_spec(scorer_obj)
    assert s.metric_fn is precision_score
    assert s.sign == 1


def test_scorer_from_spec_make_scorer_greater_is_better_false():
    """make_scorer(greater_is_better=False) -> sign=-1."""
    from sklearn.metrics import make_scorer

    from xorq.expr.ml.metrics import Scorer

    scorer_obj = make_scorer(mean_squared_error, greater_is_better=False)
    s = Scorer.from_spec(scorer_obj)
    assert s.metric_fn is mean_squared_error
    assert s.sign == -1


class TestCustomPredColName:
    """Test that custom name= on prediction methods flows through to deferred_sklearn_metric."""

    def test_classifier_custom_pred(self, classification_data):
        """Classifier: predict(name='my_pred') -> pred='my_pred'."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "cls_train")
        test_expr = api.register(test_df, "cls_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", RandomForestClassifier(n_estimators=10, random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        preds = fitted.predict(test_expr, name="my_pred")
        assert "my_pred" in preds.columns

        result = deferred_sklearn_metric(
            expr=preds,
            target="target",
            pred="my_pred",
            metric=accuracy_score,
        ).execute()
        assert 0 <= result <= 1

    def test_regressor_custom_pred(self, regression_data):
        """Regressor: predict(name='my_pred') -> pred='my_pred'."""
        train_df, test_df, feature_names = regression_data
        train_expr = api.register(train_df, "reg_train")
        test_expr = api.register(test_df, "reg_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("reg", RandomForestRegressor(n_estimators=10, random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        preds = fitted.predict(test_expr, name="my_pred")
        assert "my_pred" in preds.columns

        result = deferred_sklearn_metric(
            expr=preds,
            target="target",
            pred="my_pred",
            metric=r2_score,
        ).execute()
        assert isinstance(result, (float, np.floating))

    def test_clusterer_custom_pred(self):
        """Clusterer: predict(name='my_labels') -> pred='my_labels'."""
        KMeans = sklearn.cluster.KMeans
        adjusted_rand_score = sklearn.metrics.adjusted_rand_score

        X, y = make_classification(
            n_samples=200, n_features=5, n_informative=3, n_classes=2, random_state=42
        )
        df = pd.DataFrame(X, columns=[f"f{i}" for i in range(5)]).assign(target=y)
        train_df, test_df = train_test_split(df, test_size=0.3, random_state=42)

        train_expr = api.register(train_df, "clu_train")
        test_expr = api.register(test_df, "clu_test")

        features = [f"f{i}" for i in range(5)]
        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clu", KMeans(n_clusters=2, random_state=42, n_init=10)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=features, target="target"
        )

        preds = fitted.predict(test_expr, name="my_labels")
        assert "my_labels" in preds.columns

        result = deferred_sklearn_metric(
            expr=preds,
            target="target",
            pred="my_labels",
            metric=adjusted_rand_score,
        ).execute()
        assert isinstance(result, (float, np.floating))

    def test_predict_proba_custom_pred(self, classification_data):
        """predict_proba(name='my_proba') -> pred='my_proba'."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "proba_train")
        test_expr = api.register(test_df, "proba_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        preds = fitted.predict_proba(test_expr, name="my_proba")
        assert "my_proba" in preds.columns

        result = deferred_sklearn_metric(
            expr=preds,
            target="target",
            pred="my_proba",
            metric=roc_auc_score,
        ).execute()
        assert 0 <= result <= 1

    def test_decision_function_custom_pred(self):
        """decision_function(name='my_scores') -> pred='my_scores'."""
        X, y = make_classification(
            n_samples=200, n_features=5, n_informative=3, n_classes=2, random_state=7
        )
        df = pd.DataFrame(X, columns=[f"f{i}" for i in range(5)]).assign(target=y)
        train_df, test_df = train_test_split(df, test_size=0.3, random_state=42)

        train_expr = api.register(train_df, "df_train")
        test_expr = api.register(test_df, "df_test")

        features = [f"f{i}" for i in range(5)]
        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("svm", LinearSVC(random_state=7, max_iter=5000)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=features, target="target"
        )

        preds = fitted.decision_function(test_expr, name="my_scores")
        assert "my_scores" in preds.columns

        result = deferred_sklearn_metric(
            expr=preds,
            target="target",
            pred="my_scores",
            metric=roc_auc_score,
        ).execute()
        assert isinstance(result, (float, np.floating))


# ---------------------------------------------------------------------------
# Tests for metric_fn path (non-scorer metrics)
# ---------------------------------------------------------------------------

cohen_kappa_score = sklearn.metrics.cohen_kappa_score
hamming_loss = sklearn.metrics.hamming_loss
hinge_loss = sklearn.metrics.hinge_loss
zero_one_loss = sklearn.metrics.zero_one_loss
fbeta_score = sklearn.metrics.fbeta_score
mean_pinball_loss = sklearn.metrics.mean_pinball_loss
mean_tweedie_deviance = sklearn.metrics.mean_tweedie_deviance
d2_pinball_score = sklearn.metrics.d2_pinball_score
d2_tweedie_score = sklearn.metrics.d2_tweedie_score
d2_log_loss_score = sklearn.metrics.d2_log_loss_score
calinski_harabasz_score = sklearn.metrics.calinski_harabasz_score
davies_bouldin_score = sklearn.metrics.davies_bouldin_score
silhouette_score = sklearn.metrics.silhouette_score
coverage_error = sklearn.metrics.coverage_error
dcg_score = sklearn.metrics.dcg_score
ndcg_score = sklearn.metrics.ndcg_score
label_ranking_average_precision_score = (
    sklearn.metrics.label_ranking_average_precision_score
)
label_ranking_loss = sklearn.metrics.label_ranking_loss
class_likelihood_ratios = sklearn.metrics.class_likelihood_ratios
confusion_matrix = sklearn.metrics.confusion_matrix
homogeneity_completeness_v_measure = sklearn.metrics.homogeneity_completeness_v_measure
pair_confusion_matrix = sklearn.metrics.pair_confusion_matrix
roc_curve = sklearn.metrics.roc_curve
precision_recall_curve = sklearn.metrics.precision_recall_curve
det_curve = sklearn.metrics.det_curve
silhouette_samples = sklearn.metrics.silhouette_samples
KMeans = sklearn.cluster.KMeans


class TestMetricValidation:
    """Validate dispatch errors for the ``metric`` parameter."""

    def test_unknown_callable_raises(self):
        """Unknown callable -> ValueError."""

        def my_custom_fn(y_true, y_pred):
            return 0.0

        df = pd.DataFrame({"target": [0, 1], "predict": [0, 1]})
        expr = api.register(df, "unknown_callable")
        with pytest.raises(ValueError, match="Unknown callable"):
            deferred_sklearn_metric(
                expr=expr,
                target="target",
                pred="predict",
                metric=my_custom_fn,
            )

    def test_invalid_type_raises(self):
        """Non-callable, non-string -> TypeError."""
        df = pd.DataFrame({"target": [0, 1], "predict": [0, 1]})
        expr = api.register(df, "invalid_type")
        with pytest.raises(TypeError, match="metric must be"):
            deferred_sklearn_metric(
                expr=expr,
                target="target",
                pred="predict",
                metric=42,
            )


class TestMetricFnYPred:
    """Non-scorer metrics that take (y_true, y_pred) -> scalar float."""

    @pytest.fixture
    def classification_expr(self, classification_data):
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "mfn_train")
        test_expr = api.register(test_df, "mfn_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                (
                    "classifier",
                    RandomForestClassifier(n_estimators=10, random_state=42),
                ),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_pred = sklearn_pipeline.predict(test_df[feature_names])

        return (
            fitted.predict(test_expr),
            test_df["target"].values,
            y_pred,
        )

    @pytest.mark.parametrize(
        "metric_func,kwargs",
        (
            (cohen_kappa_score, {}),
            (hamming_loss, {}),
            (zero_one_loss, {}),
            (fbeta_score, {"beta": 1.0}),
            (mean_pinball_loss, {}),
            (d2_pinball_score, {}),
            (mean_tweedie_deviance, {}),
            (d2_tweedie_score, {}),
        ),
        ids=lambda p: p.__name__ if callable(p) else str(p),
    )
    def test_y_pred_metric(self, classification_expr, metric_func, kwargs):
        expr_with_preds, y_true, y_pred = classification_expr

        result = deferred_sklearn_metric(
            expr=expr_with_preds,
            target="target",
            pred=ResponseMethod.PREDICT,
            metric=metric_func,
            metric_kwargs=kwargs,
        ).execute()

        expected = metric_func(y_true, y_pred, **kwargs)
        assert abs(result - expected) < 1e-10


class TestMetricFnYScore:
    """Non-scorer metrics that take (y_true, y_score/pred_decision)."""

    def test_hinge_loss(self, classification_data):
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "hinge_train")
        test_expr = api.register(test_df, "hinge_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("svm", LinearSVC(random_state=42, max_iter=5000)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_scores = fitted.decision_function(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_scores,
            target="target",
            pred=ResponseMethod.DECISION_FUNCTION,
            metric=hinge_loss,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_score = sklearn_pipeline.decision_function(test_df[feature_names])
        expected = hinge_loss(test_df["target"], y_score)
        assert abs(result - expected) < 1e-10

    def test_d2_log_loss_score(self, classification_data):
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "d2ll_train")
        test_expr = api.register(test_df, "d2ll_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_proba = fitted.predict_proba(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=d2_log_loss_score,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_proba = sklearn_pipeline.predict_proba(test_df[feature_names])[:, 1]
        expected = d2_log_loss_score(test_df["target"], y_proba)
        assert abs(result - expected) < 1e-10


class TestMetricFnClustering:
    """Clustering metrics that take (X, labels) via tuple target."""

    @pytest.fixture
    def clustering_expr(self):
        from sklearn.datasets import make_blobs

        X, _ = make_blobs(n_samples=200, centers=3, n_features=4, random_state=42)
        feature_names = tuple(f"f{i}" for i in range(X.shape[1]))
        df = pd.DataFrame(X, columns=list(feature_names))
        kmeans = KMeans(n_clusters=3, random_state=42, n_init=10)
        df["cluster"] = kmeans.fit_predict(X)
        expr = api.register(df, "cluster_test")
        return expr, feature_names, X, df["cluster"].values

    @pytest.mark.parametrize(
        "metric_func",
        (
            calinski_harabasz_score,
            davies_bouldin_score,
            silhouette_score,
        ),
        ids=lambda p: p.__name__,
    )
    def test_clustering_metric(self, clustering_expr, metric_func):
        expr, feature_names, X, labels = clustering_expr

        result = deferred_sklearn_metric(
            expr=expr,
            target=feature_names,
            pred="cluster",
            metric=metric_func,
        ).execute()

        expected = metric_func(X, labels)
        assert abs(result - expected) < 1e-10


class TestMetricFnSignAutoDetect:
    """When metric_fn is a known scorer func, sign is auto-detected."""

    def test_known_scorer_fn_via_metric_fn_gets_sign(self):
        """mean_squared_error is in _build_known_scorer_funcs;
        Scorer.from_spec resolves it with sign=1 (bare callable convention).
        metric_fn path picks up that same sign=1."""
        df = pd.DataFrame(
            {
                "target": [0.0, 1.0, 2.0, 3.0],
                "predict": [0.1, 0.9, 2.1, 3.2],
            }
        )
        expr = api.register(df, "sign_auto_test")

        result_metric_fn = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=mean_squared_error,
        ).execute()

        # Bare callable through scorer path also uses sign=1
        result_scorer = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=mean_squared_error,
        ).execute()

        assert abs(result_metric_fn - result_scorer) < 1e-10

        # neg_mean_squared_error string scorer uses sign=-1
        result_neg = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric="neg_mean_squared_error",
        ).execute()

        assert abs(result_metric_fn + result_neg) < 1e-10

    def test_non_scorer_fn_has_no_sign(self):
        df = pd.DataFrame(
            {
                "target": [0, 1, 0, 1],
                "predict": [0, 1, 1, 1],
            }
        )
        expr = api.register(df, "no_sign_test")

        result = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=cohen_kappa_score,
        ).execute()

        expected = cohen_kappa_score([0, 1, 0, 1], [0, 1, 1, 1])
        assert abs(result - expected) < 1e-10


class TestMetricFnMultilabel:
    """Multilabel/ranking metrics that take (y_true_2D, y_score_2D).

    Both arguments are 2D matrices. target is a tuple of label column names;
    pred is a single array column whose rows are reconstructed into a
    2D ndarray via _reconstruct_matrix (no positive-class extraction).
    """

    @pytest.fixture
    def multilabel_expr(self):
        from sklearn.multiclass import OneVsRestClassifier

        rng = np.random.RandomState(42)
        X = rng.randn(100, 5)
        Y = rng.randint(0, 2, size=(100, 3))

        clf = OneVsRestClassifier(LogisticRegression(random_state=42))
        clf.fit(X[:70], Y[:70])

        y_true = Y[70:]
        y_score = clf.predict_proba(X[70:])

        label_names = tuple(f"label_{i}" for i in range(Y.shape[1]))
        df = pd.DataFrame(y_true, columns=list(label_names))
        # Store y_score as a single array column (mimics predict_proba storage)
        df["predict_proba"] = list(y_score)

        expr = api.register(df, "multilabel_test")
        return expr, label_names, y_true, y_score

    @pytest.mark.parametrize(
        "metric_func",
        (
            coverage_error,
            dcg_score,
            ndcg_score,
            label_ranking_average_precision_score,
            label_ranking_loss,
        ),
        ids=lambda p: p.__name__,
    )
    def test_multilabel_metric(self, multilabel_expr, metric_func):
        expr, label_names, y_true, y_score = multilabel_expr

        result = deferred_sklearn_metric(
            expr=expr,
            target=label_names,
            pred="predict_proba",
            metric=metric_func,
        ).execute()

        expected = metric_func(y_true, y_score)
        assert abs(result - expected) < 1e-10

    @pytest.mark.parametrize(
        "metric_func",
        (
            coverage_error,
            dcg_score,
            ndcg_score,
            label_ranking_average_precision_score,
            label_ranking_loss,
        ),
        ids=lambda p: p.__name__,
    )
    def test_multilabel_metric_tuple_pred(self, metric_func):
        """Scores stored as separate columns; pred is a tuple of str."""
        from sklearn.multiclass import OneVsRestClassifier

        rng = np.random.RandomState(42)
        X = rng.randn(100, 5)
        Y = rng.randint(0, 2, size=(100, 3))

        clf = OneVsRestClassifier(LogisticRegression(random_state=42))
        clf.fit(X[:70], Y[:70])

        y_true = Y[70:]
        y_score = clf.predict_proba(X[70:])

        label_names = tuple(f"label_{i}" for i in range(Y.shape[1]))
        score_names = tuple(f"score_{i}" for i in range(y_score.shape[1]))
        df = pd.DataFrame(y_true, columns=list(label_names))
        for i, name in enumerate(score_names):
            df[name] = y_score[:, i]

        expr = api.register(df, "multilabel_tuple_pred")

        result = deferred_sklearn_metric(
            expr=expr,
            target=label_names,
            pred=score_names,
            metric=metric_func,
        ).execute()

        expected = metric_func(y_true, y_score)
        assert abs(result - expected) < 1e-10


# ---------------------------------------------------------------------------
# Tests for non-scalar return metrics (auto-detected via registry)
# ---------------------------------------------------------------------------


class TestNonScalarMetrics:
    """Metrics that return structs, arrays, or per-sample values.

    return_type is auto-resolved from _build_metric_return_types().
    """

    # -- tuple of scalars -> Struct -------------------------------------------

    def test_class_likelihood_ratios(self, classification_data):
        """class_likelihood_ratios -> Struct(positive_lr, negative_lr)."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "clr_train")
        test_expr = api.register(test_df, "clr_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", RandomForestClassifier(n_estimators=10, random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_preds = fitted.predict(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_preds,
            target="target",
            pred=ResponseMethod.PREDICT,
            metric=class_likelihood_ratios,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_pred = sklearn_pipeline.predict(test_df[feature_names])
        expected = class_likelihood_ratios(test_df["target"], y_pred)

        assert isinstance(result, dict)
        assert set(result.keys()) == {
            "positive_likelihood_ratio",
            "negative_likelihood_ratio",
        }
        assert np.isclose(result["positive_likelihood_ratio"], expected[0])
        assert np.isclose(result["negative_likelihood_ratio"], expected[1])

    def test_homogeneity_completeness_v_measure(self):
        """homogeneity_completeness_v_measure -> Struct(h, c, v)."""
        y_true = [0, 0, 0, 1, 1, 1, 2, 2, 2]
        y_pred = [0, 0, 1, 1, 1, 0, 2, 2, 2]
        df = pd.DataFrame({"target": y_true, "predict": y_pred})
        expr = api.register(df, "hcv_test")

        result = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=homogeneity_completeness_v_measure,
        ).execute()

        expected = homogeneity_completeness_v_measure(y_true, y_pred)

        assert isinstance(result, dict)
        assert set(result.keys()) == {"homogeneity", "completeness", "v_measure"}
        assert np.isclose(result["homogeneity"], expected[0])
        assert np.isclose(result["completeness"], expected[1])
        assert np.isclose(result["v_measure"], expected[2])

    # -- matrix -> Array(Array(int64)) ----------------------------------------

    def test_confusion_matrix(self, classification_data):
        """confusion_matrix -> Array(Array(int64))."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "cm_train")
        test_expr = api.register(test_df, "cm_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", RandomForestClassifier(n_estimators=10, random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_preds = fitted.predict(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_preds,
            target="target",
            pred=ResponseMethod.PREDICT,
            metric=confusion_matrix,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_pred = sklearn_pipeline.predict(test_df[feature_names])
        expected = confusion_matrix(test_df["target"], y_pred)

        assert isinstance(result, list)
        assert np.array_equal(np.array(result), expected)

    def test_pair_confusion_matrix(self):
        """pair_confusion_matrix -> Array(Array(int64))."""
        y_true = [0, 0, 1, 1, 2, 2]
        y_pred = [0, 0, 1, 2, 2, 2]
        df = pd.DataFrame({"target": y_true, "predict": y_pred})
        expr = api.register(df, "pcm_test")

        result = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=pair_confusion_matrix,
        ).execute()

        expected = pair_confusion_matrix(y_true, y_pred)

        assert isinstance(result, list)
        assert np.array_equal(np.array(result), expected)

    # -- curves -> Struct of arrays -------------------------------------------

    def test_roc_curve(self, classification_data):
        """roc_curve -> Struct(fpr, tpr, thresholds)."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "roc_train")
        test_expr = api.register(test_df, "roc_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_proba = fitted.predict_proba(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=roc_curve,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_proba = sklearn_pipeline.predict_proba(test_df[feature_names])[:, 1]
        expected = roc_curve(test_df["target"], y_proba)

        assert isinstance(result, dict)
        assert set(result.keys()) == {"fpr", "tpr", "thresholds"}
        assert np.allclose(result["fpr"], expected[0])
        assert np.allclose(result["tpr"], expected[1])
        assert np.allclose(result["thresholds"], expected[2])

    def test_precision_recall_curve(self, classification_data):
        """precision_recall_curve -> Struct(precision, recall, thresholds)."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "prc_train")
        test_expr = api.register(test_df, "prc_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_proba = fitted.predict_proba(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=precision_recall_curve,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_proba = sklearn_pipeline.predict_proba(test_df[feature_names])[:, 1]
        expected = precision_recall_curve(test_df["target"], y_proba)

        assert isinstance(result, dict)
        assert set(result.keys()) == {"precision", "recall", "thresholds"}
        assert np.allclose(result["precision"], expected[0])
        assert np.allclose(result["recall"], expected[1])
        assert np.allclose(result["thresholds"], expected[2])

    def test_det_curve(self, classification_data):
        """det_curve -> Struct(fpr, fnr, thresholds)."""
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "det_train")
        test_expr = api.register(test_df, "det_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )

        expr_with_proba = fitted.predict_proba(test_expr)
        result = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=det_curve,
        ).execute()

        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_proba = sklearn_pipeline.predict_proba(test_df[feature_names])[:, 1]
        expected = det_curve(test_df["target"], y_proba)

        assert isinstance(result, dict)
        assert set(result.keys()) == {"fpr", "fnr", "thresholds"}
        assert np.allclose(result["fpr"], expected[0])
        assert np.allclose(result["fnr"], expected[1])
        assert np.allclose(result["thresholds"], expected[2])

    # -- per-sample -> Array(float64) -----------------------------------------

    def test_silhouette_samples(self):
        """silhouette_samples -> Array(float64)."""
        from sklearn.datasets import make_blobs

        X, _ = make_blobs(n_samples=50, centers=3, n_features=4, random_state=42)
        labels = KMeans(n_clusters=3, random_state=42, n_init=10).fit_predict(X)

        feature_names = tuple(f"f{i}" for i in range(X.shape[1]))
        df = pd.DataFrame(X, columns=list(feature_names))
        df["cluster"] = labels
        expr = api.register(df, "silhouette_samples_test")

        result = deferred_sklearn_metric(
            expr=expr,
            target=feature_names,
            pred="cluster",
            metric=silhouette_samples,
        ).execute()

        expected = silhouette_samples(X, labels)

        assert isinstance(result, list)
        assert np.allclose(result, expected)


class TestDeferredAucFromCurve:
    """Tests for deferred_auc_from_curve — auc computed from deferred curves."""

    @pytest.fixture
    def proba_expr(self, classification_data):
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "auc_train")
        test_expr = api.register(test_df, "auc_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )
        expr_with_proba = fitted.predict_proba(test_expr)

        # Also compute sklearn reference
        sklearn_pipeline.fit(train_df[feature_names], train_df["target"])
        y_proba = sklearn_pipeline.predict_proba(test_df[feature_names])[:, 1]
        y_true = test_df["target"]
        return expr_with_proba, y_true, y_proba

    def test_roc_auc(self, proba_expr):
        from sklearn.metrics import auc, roc_curve

        from xorq.expr.ml.metrics import deferred_auc_from_curve

        expr_with_proba, y_true, y_proba = proba_expr
        deferred_roc = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=roc_curve,
        )
        result = deferred_auc_from_curve(deferred_roc).execute()

        fpr, tpr, _ = roc_curve(y_true, y_proba)
        expected = auc(fpr, tpr)
        assert abs(result - expected) < 1e-10

    def test_precision_recall_auc(self, proba_expr):
        from sklearn.metrics import auc, precision_recall_curve

        from xorq.expr.ml.metrics import deferred_auc_from_curve

        expr_with_proba, y_true, y_proba = proba_expr
        deferred_pr = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=precision_recall_curve,
        )
        result = deferred_auc_from_curve(deferred_pr).execute()

        precision, recall, _ = precision_recall_curve(y_true, y_proba)
        expected = auc(recall, precision)
        assert abs(result - expected) < 1e-10

    def test_det_auc(self, proba_expr):
        from sklearn.metrics import auc, det_curve

        from xorq.expr.ml.metrics import deferred_auc_from_curve

        expr_with_proba, y_true, y_proba = proba_expr
        deferred_det = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=det_curve,
        )
        result = deferred_auc_from_curve(deferred_det).execute()

        fpr, fnr, _ = det_curve(y_true, y_proba)
        expected = auc(fpr, fnr)
        assert abs(result - expected) < 1e-10

    def test_compose_into_table(self, proba_expr):
        """auc composes with other metrics via .as_scalar() + .mutate()."""
        from sklearn.metrics import roc_curve

        from xorq.expr.ml.metrics import deferred_auc_from_curve

        expr_with_proba, _, _ = proba_expr
        deferred_roc = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=roc_curve,
        )
        deferred_roc_auc = deferred_auc_from_curve(deferred_roc)
        deferred_roc_auc_score = deferred_sklearn_metric(
            expr=expr_with_proba,
            target="target",
            pred=ResponseMethod.PREDICT_PROBA,
            metric=roc_auc_score,
        )

        table = (
            deferred_roc_auc_score.as_scalar()
            .name("roc_auc_score")
            .as_table()
            .mutate(roc_auc=deferred_roc_auc.as_scalar())
        )
        result = table.execute()
        assert "roc_auc_score" in result.columns
        assert "roc_auc" in result.columns
        assert len(result) == 1

    def test_invalid_input_non_struct(self):
        """Raises TypeError for non-Struct expressions."""
        from xorq.expr.ml.metrics import deferred_auc_from_curve

        df = pd.DataFrame({"a": [1, 2, 3], "target": [0, 1, 0]})
        expr = api.register(df, "auc_invalid")
        deferred_acc = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="a",
            metric=accuracy_score,
        )
        with pytest.raises(TypeError, match="Expected a Struct"):
            deferred_auc_from_curve(deferred_acc)

    def test_unrecognized_curve_fields(self):
        """Raises ValueError for Struct with unrecognized field names."""
        from unittest.mock import MagicMock

        from xorq.expr.ml.metrics import deferred_auc_from_curve

        # Mock an expression whose .type() returns a dt.Struct with wrong field names
        fake_struct_type = dt.Struct(
            {
                "x": dt.Array(dt.float64),
                "y": dt.Array(dt.float64),
                "z": dt.Array(dt.float64),
            }
        )
        mock_expr = MagicMock()
        mock_expr.type.return_value = fake_struct_type
        with pytest.raises(ValueError, match="Unrecognized curve fields"):
            deferred_auc_from_curve(mock_expr)


# ---------------------------------------------------------------------------
# Tests for error branches / edge cases to improve coverage
# ---------------------------------------------------------------------------


class TestDefaultScorerForModel:
    """Cover _default_scorer_for_model error branch (metrics.py:69-70)."""

    def test_unknown_model_type_raises(self):
        """Non-classifier/regressor/clusterer -> ValueError."""
        from sklearn.base import BaseEstimator

        from xorq.expr.ml.metrics import _default_scorer_for_model

        class UnknownEstimator(BaseEstimator):
            pass

        model = UnknownEstimator()
        with pytest.raises(ValueError, match="Cannot determine default scorer"):
            _default_scorer_for_model(model)


class TestScorerResolveResponseMethod:
    """Cover Scorer._resolve_response_method fallback (metrics.py:172-173)."""

    def test_unexpected_response_method_raises(self):
        """Non-string, non-tuple _response_method -> ValueError."""
        from unittest.mock import MagicMock

        from xorq.expr.ml.metrics import Scorer

        mock_scorer = MagicMock()
        mock_scorer._response_method = 42  # neither str nor tuple
        with pytest.raises(ValueError, match="Unexpected _response_method"):
            Scorer._resolve_response_method(mock_scorer)


class TestDeferredSklearnMetricScorerInput:
    """Cover deferred_sklearn_metric with Scorer input directly (metrics.py:596-597)."""

    def test_scorer_instance_input(self):
        """Passing a Scorer instance directly to deferred_sklearn_metric."""
        from xorq.expr.ml.metrics import Scorer

        df = pd.DataFrame(
            {
                "target": [0, 1, 0, 1, 1],
                "predict": [0, 1, 1, 1, 0],
            }
        )
        expr = api.register(df, "scorer_direct")

        scorer = Scorer(
            metric_fn=accuracy_score,
            sign=1,
            kwargs={},
            response_method=ResponseMethod.PREDICT,
        )
        result = deferred_sklearn_metric(
            expr=expr,
            target="target",
            pred="predict",
            metric=scorer,
        ).execute()

        expected = accuracy_score([0, 1, 0, 1, 1], [0, 1, 1, 1, 0])
        assert np.isclose(result, expected)


class TestPipelineScoreMethods:
    """Cover FittedPipeline.score_expr() and .score() (pipeline_lib.py:953-1026)."""

    @pytest.fixture
    def fitted_classifier_pipeline(self, classification_data):
        train_df, test_df, feature_names = classification_data
        train_expr = api.register(train_df, "score_train")
        test_expr = api.register(test_df, "score_test")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("clf", LogisticRegression(random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )
        return fitted, test_expr, train_df, test_df, feature_names

    def test_score_expr_default(self, fitted_classifier_pipeline):
        """score_expr with default scorer (accuracy for classifier)."""
        fitted, test_expr, _, test_df, feature_names = fitted_classifier_pipeline
        result = fitted.score_expr(test_expr).execute()
        assert isinstance(result, (float, np.floating))
        assert 0 <= result <= 1

    def test_score_expr_with_string_scorer(self, fitted_classifier_pipeline):
        """score_expr with string scorer name."""
        fitted, test_expr, _, _, _ = fitted_classifier_pipeline
        result = fitted.score_expr(test_expr, scorer="accuracy").execute()
        assert isinstance(result, (float, np.floating))
        assert 0 <= result <= 1

    def test_score_default(self, fitted_classifier_pipeline):
        """score() with default scorer using numpy arrays."""
        fitted, _, _, test_df, feature_names = fitted_classifier_pipeline
        X_test = test_df[feature_names].values
        y_test = test_df["target"].values
        result = fitted.score(X_test, y_test)
        assert isinstance(result, (float, np.floating))
        assert 0 <= result <= 1

    def test_score_with_scorer(self, fitted_classifier_pipeline):
        """score() with explicit scorer string."""
        fitted, _, _, test_df, feature_names = fitted_classifier_pipeline
        X_test = test_df[feature_names].values
        y_test = test_df["target"].values
        result = fitted.score(X_test, y_test, scorer="accuracy")
        assert isinstance(result, (float, np.floating))
        assert 0 <= result <= 1

    def test_score_no_predict_step_raises(self):
        """score() on a transform-only pipeline -> ValueError."""

        df = pd.DataFrame({"feature_0": range(20), "feature_1": range(20)})
        train_expr = api.register(df, "transform_only_train")

        sklearn_pipeline = SkPipeline([("scaler", StandardScaler())])
        xorq_pipeline = Pipeline.from_instance(sklearn_pipeline)
        fitted = xorq_pipeline.fit(train_expr, features=["feature_0", "feature_1"])

        with pytest.raises(ValueError, match="Pipeline does not have a predict step"):
            fitted.score(
                df[["feature_0", "feature_1"]].values,
                np.zeros(20),
            )

    def test_score_regressor(self, regression_data):
        """score() with a regressor pipeline uses default r2 scorer."""
        train_df, test_df, feature_names = regression_data
        train_expr = api.register(train_df, "score_reg_train")

        sklearn_pipeline = SkPipeline(
            [
                ("scaler", StandardScaler()),
                ("reg", RandomForestRegressor(n_estimators=10, random_state=42)),
            ]
        )
        fitted = Pipeline.from_instance(sklearn_pipeline).fit(
            train_expr, features=feature_names, target="target"
        )
        X_test = test_df[feature_names].values
        y_test = test_df["target"].values
        result = fitted.score(X_test, y_test)
        assert isinstance(result, (float, np.floating))
